=== modified file 'drizzled/cursor.cc'
--- drizzled/cursor.cc	2010-02-18 16:19:37 +0000
+++ drizzled/cursor.cc	2010-02-19 16:11:03 +0000
@@ -696,9 +696,9 @@
 
 inline
 void
-Cursor::mark_trx_read_write()
+Cursor::setTransactionReadWrite()
 {
-  Ha_trx_info *ha_info= ha_session()->getEngineInfo(engine);
+  ResourceContext *resource_context= ha_session()->getResourceContext(engine);
   /*
     When a storage engine method is called, the transaction must
     have been started, unless it's a DDL call, for which the
@@ -707,9 +707,9 @@
     Unfortunately here we can't know know for sure if the engine
     has registered the transaction or not, so we must check.
   */
-  if (ha_info->is_started())
+  if (resource_context->isStarted())
   {
-      ha_info->set_trx_read_write();
+    resource_context->markModifiedData();
   }
 }
 
@@ -727,7 +727,7 @@
 int
 Cursor::ha_delete_all_rows()
 {
-  mark_trx_read_write();
+  setTransactionReadWrite();
 
   int result= delete_all_rows();
 
@@ -757,7 +757,7 @@
 int
 Cursor::ha_reset_auto_increment(uint64_t value)
 {
-  mark_trx_read_write();
+  setTransactionReadWrite();
 
   return reset_auto_increment(value);
 }
@@ -772,7 +772,7 @@
 int
 Cursor::ha_analyze(Session* session, HA_CHECK_OPT*)
 {
-  mark_trx_read_write();
+  setTransactionReadWrite();
 
   return analyze(session);
 }
@@ -786,7 +786,7 @@
 int
 Cursor::ha_disable_indexes(uint32_t mode)
 {
-  mark_trx_read_write();
+  setTransactionReadWrite();
 
   return disable_indexes(mode);
 }
@@ -801,7 +801,7 @@
 int
 Cursor::ha_enable_indexes(uint32_t mode)
 {
-  mark_trx_read_write();
+  setTransactionReadWrite();
 
   return enable_indexes(mode);
 }
@@ -816,7 +816,7 @@
 int
 Cursor::ha_discard_or_import_tablespace(bool discard)
 {
-  mark_trx_read_write();
+  setTransactionReadWrite();
 
   return discard_or_import_tablespace(discard);
 }
@@ -830,7 +830,7 @@
 void
 Cursor::closeMarkForDelete(const char *name)
 {
-  mark_trx_read_write();
+  setTransactionReadWrite();
 
   return drop_table(name);
 }
@@ -1513,7 +1513,7 @@
     table->timestamp_field->set_time();
 
   DRIZZLE_INSERT_ROW_START(table_share->db.str, table_share->table_name.str);
-  mark_trx_read_write();
+  setTransactionReadWrite();
   error= write_row(buf);
   DRIZZLE_INSERT_ROW_DONE(error);
 
@@ -1540,7 +1540,7 @@
   assert(new_data == table->record[0]);
 
   DRIZZLE_UPDATE_ROW_START(table_share->db.str, table_share->table_name.str);
-  mark_trx_read_write();
+  setTransactionReadWrite();
   error= update_row(old_data, new_data);
   DRIZZLE_UPDATE_ROW_DONE(error);
 
@@ -1560,7 +1560,7 @@
   int error;
 
   DRIZZLE_DELETE_ROW_START(table_share->db.str, table_share->table_name.str);
-  mark_trx_read_write();
+  setTransactionReadWrite();
   error= delete_row(buf);
   DRIZZLE_DELETE_ROW_DONE(error);
 

=== modified file 'drizzled/cursor.h'
--- drizzled/cursor.h	2010-02-18 16:19:37 +0000
+++ drizzled/cursor.h	2010-02-19 16:11:03 +0000
@@ -555,7 +555,7 @@
 
 private:
   /* Private helpers */
-  inline void mark_trx_read_write();
+  inline void setTransactionReadWrite();
 private:
   /*
     Low-level primitives for storage engines.  These should be

=== modified file 'drizzled/drizzled.cc'
--- drizzled/drizzled.cc	2010-02-18 18:47:12 +0000
+++ drizzled/drizzled.cc	2010-02-19 16:11:03 +0000
@@ -47,6 +47,7 @@
 #include "drizzled/plugin/error_message.h"
 #include "drizzled/plugin/client.h"
 #include "drizzled/plugin/scheduler.h"
+#include "drizzled/plugin/xa_storage_engine.h"
 #include "drizzled/probes.h"
 #include "drizzled/session_list.h"
 #include "drizzled/charset.h"
@@ -1378,7 +1379,7 @@
     }
   }
 
-  if (plugin::StorageEngine::recover(0))
+  if (plugin::XaStorageEngine::recoverAllXids(0))
   {
     unireg_abort(1);
   }

=== modified file 'drizzled/filesort.cc'
--- drizzled/filesort.cc	2010-02-04 08:14:46 +0000
+++ drizzled/filesort.cc	2010-02-14 19:27:57 +0000
@@ -40,6 +40,7 @@
 #include "drizzled/internal/iocache.h"
 #include "drizzled/internal/my_sys.h"
 #include "plugin/myisam/myisam.h"
+#include "drizzled/plugin/transactional_storage_engine.h"
 
 using namespace std;
 
@@ -153,7 +154,7 @@
    Release InnoDB's adaptive hash index latch (if holding) before
    running a sort.
   */
-  plugin::StorageEngine::releaseTemporaryLatches(session);
+  plugin::TransactionalStorageEngine::releaseTemporaryLatches(session);
 
   /*
     Don't use table->sort in filesort as it is also used by

=== modified file 'drizzled/handler_structs.h'
--- drizzled/handler_structs.h	2010-02-04 08:14:46 +0000
+++ drizzled/handler_structs.h	2010-02-06 01:51:20 +0000
@@ -40,7 +40,6 @@
 namespace drizzled
 {
 
-class Ha_trx_info;
 struct st_key;
 typedef struct st_key KEY;
 typedef struct st_key_cache KEY_CACHE;
@@ -51,51 +50,6 @@
 class StorageEngine;
 }
 
-struct Session_TRANS
-{
-  Session_TRANS() :
-    no_2pc(false),
-    ha_list(NULL),
-    modified_non_trans_table(false)
-  {}
-
-  /* true is not all entries in the engines[] support 2pc */
-  bool        no_2pc;
-  /* storage engines that registered in this transaction */
-  Ha_trx_info *ha_list;
-  /*
-    The purpose of this flag is to keep track of non-transactional
-    tables that were modified in scope of:
-    - transaction, when the variable is a member of
-    Session::transaction.all
-    - top-level statement or sub-statement, when the variable is a
-    member of Session::transaction.stmt
-    This member has the following life cycle:
-    * stmt.modified_non_trans_table is used to keep track of
-    modified non-transactional tables of top-level statements. At
-    the end of the previous statement and at the beginning of the session,
-    it is reset to false.  If such functions
-    as mysql_insert, mysql_update, mysql_delete etc modify a
-    non-transactional table, they set this flag to true.  At the
-    end of the statement, the value of stmt.modified_non_trans_table
-    is merged with all.modified_non_trans_table and gets reset.
-    * all.modified_non_trans_table is reset at the end of transaction
-
-    * Since we do not have a dedicated context for execution of a
-    sub-statement, to keep track of non-transactional changes in a
-    sub-statement, we re-use stmt.modified_non_trans_table.
-    At entrance into a sub-statement, a copy of the value of
-    stmt.modified_non_trans_table (containing the changes of the
-    outer statement) is saved on stack. Then
-    stmt.modified_non_trans_table is reset to false and the
-    substatement is executed. Then the new value is merged with the
-    saved value.
-  */
-  bool modified_non_trans_table;
-
-  void reset() { no_2pc= false; modified_non_trans_table= false; }
-};
-
 typedef struct st_ha_create_information
 {
   const CHARSET_INFO *table_charset, *default_table_charset;

=== modified file 'drizzled/include.am'
--- drizzled/include.am	2010-02-18 16:16:20 +0000
+++ drizzled/include.am	2010-02-19 16:11:03 +0000
@@ -198,7 +198,6 @@
 			      drizzled/gettext.h \
 			      drizzled/global_charset_info.h \
 			      drizzled/ha_statistics.h \
-			      drizzled/ha_trx_info.h \
 			      drizzled/cursor.h \
 			      drizzled/handler_structs.h \
 			      drizzled/hash.h \
@@ -314,6 +313,8 @@
 			      drizzled/plugin/transaction_applier.h \
 			      drizzled/plugin/transaction_reader.h \
 			      drizzled/plugin/transaction_replicator.h \
+			      drizzled/plugin/transactional_storage_engine.h \
+			      drizzled/plugin/xa_storage_engine.h \
 			      drizzled/plugin/version.h \
 			      drizzled/pthread_globals.h \
 			      drizzled/probes.h \
@@ -321,6 +322,7 @@
 			      drizzled/query_id.h \
 			      drizzled/records.h \
 			      drizzled/replication_services.h \
+			      drizzled/resource_context.h \
 			      drizzled/security_context.h \
 			      drizzled/select_create.h \
 			      drizzled/select_dump.h \
@@ -412,6 +414,7 @@
 			      drizzled/time_functions.h \
 			      drizzled/tmp_table_param.h \
 			      drizzled/transaction_services.h \
+			      drizzled/transaction_context.h \
 			      drizzled/tzfile.h \
 			      drizzled/tztime.h \
 			      drizzled/unique.h \
@@ -590,7 +593,6 @@
 			   drizzled/function/units.cc \
 			   drizzled/function/user_var_as_out_param.cc \
 			   drizzled/ha_commands.cc \
-			   drizzled/ha_trx_info.cc \
 			   drizzled/hybrid_type_traits.cc \
 			   drizzled/hybrid_type_traits_decimal.cc \
 			   drizzled/hybrid_type_traits_integer.cc \
@@ -671,9 +673,12 @@
 			   drizzled/plugin/table_function.cc \
 			   drizzled/plugin/transaction_applier.cc \
 			   drizzled/plugin/transaction_replicator.cc \
+			   drizzled/plugin/transactional_storage_engine.cc \
+			   drizzled/plugin/xa_storage_engine.cc \
 			   drizzled/query_id.cc \
 			   drizzled/records.cc \
 			   drizzled/replication_services.cc \
+			   drizzled/resource_context.cc \
 			   drizzled/session.cc \
 			   drizzled/session_list.cc \
 			   drizzled/set_var.cc \

=== modified file 'drizzled/named_savepoint.h'
--- drizzled/named_savepoint.h	2010-01-28 22:19:19 +0000
+++ drizzled/named_savepoint.h	2010-02-04 05:56:17 +0000
@@ -24,7 +24,7 @@
 #ifndef DRIZZLED_NAMED_SAVEPOINT_H
 #define DRIZZLED_NAMED_SAVEPOINT_H
 
-class Ha_trx_info;
+#include "drizzled/transaction_context.h" /* for TransactionContext::ResourceContexts */
 
 namespace drizzled
 {
@@ -40,12 +40,24 @@
    * Constructor
    */
   NamedSavepoint(const char *in_name, size_t in_name_length) :
-    ha_list(NULL),
-    name(in_name, in_name_length)
+    name(in_name, in_name_length),
+    resource_contexts()
   {}
   ~NamedSavepoint()
   {}
-  Ha_trx_info *ha_list;
+
+  void setResourceContexts(TransactionContext::ResourceContexts &new_contexts)
+  {
+    resource_contexts.assign(new_contexts.begin(), new_contexts.end());
+  }
+  const TransactionContext::ResourceContexts &getResourceContexts() const
+  {
+    return resource_contexts;
+  }
+  TransactionContext::ResourceContexts &getResourceContexts()
+  {
+    return resource_contexts;
+  }
   const std::string &getName() const
   {
     return name;
@@ -57,10 +69,24 @@
   NamedSavepoint(const NamedSavepoint &other)
   {
     name.assign(other.getName());
-    ha_list= other.ha_list;
+    const TransactionContext::ResourceContexts &other_resource_contexts= other.getResourceContexts();
+    resource_contexts.assign(other_resource_contexts.begin(),
+                             other_resource_contexts.end());
+  }
+  NamedSavepoint &operator=(const NamedSavepoint &other)
+  {
+    if (this == &other)
+      return *this;
+
+    name.assign(other.getName());
+    const TransactionContext::ResourceContexts &other_resource_contexts= other.getResourceContexts();
+    resource_contexts.assign(other_resource_contexts.begin(),
+                             other_resource_contexts.end());
+    return *this;
   }
 private:
   std::string name;
+  TransactionContext::ResourceContexts resource_contexts;
   NamedSavepoint();
 };
 

=== modified file 'drizzled/plugin/storage_engine.cc'
--- drizzled/plugin/storage_engine.cc	2010-02-11 21:53:40 +0000
+++ drizzled/plugin/storage_engine.cc	2010-02-15 20:29:01 +0000
@@ -50,11 +50,8 @@
 #include "drizzled/internal/my_sys.h"
 #include "drizzled/db.h"
 
-
 #include <drizzled/table_proto.h>
 
-#include "drizzled/hash.h"
-
 static bool shutdown_has_begun= false; // Once we put in the container for the vector/etc for engines this will go away.
 
 using namespace std;
@@ -62,22 +59,19 @@
 namespace drizzled
 {
 
-typedef hash_map<std::string, plugin::StorageEngine *> EngineMap;
-typedef std::vector<plugin::StorageEngine *> EngineVector;
+namespace plugin
+{
 
 static EngineVector vector_of_engines;
-static EngineVector vector_of_transactional_engines;
 
-const std::string plugin::UNKNOWN_STRING("UNKNOWN");
-const std::string plugin::DEFAULT_DEFINITION_FILE_EXT(".dfe");
+const std::string UNKNOWN_STRING("UNKNOWN");
+const std::string DEFAULT_DEFINITION_FILE_EXT(".dfe");
 
 static std::set<std::string> set_of_table_definition_ext;
 
-plugin::StorageEngine::StorageEngine(const string name_arg,
-                                     const bitset<HTON_BIT_SIZE> &flags_arg,
-                                     bool support_2pc)
+StorageEngine::StorageEngine(const string name_arg,
+                                     const bitset<HTON_BIT_SIZE> &flags_arg)
     : Plugin(name_arg, "StorageEngine"),
-      two_phase_commit(support_2pc),
       enabled(true),
       flags(flags_arg),
       slot(0)
@@ -85,42 +79,22 @@
   if (enabled)
   {
     slot= total_ha++;
-    if (two_phase_commit)
-        total_ha_2pc++;
   }
   pthread_mutex_init(&proto_cache_mutex, NULL);
 }
 
-
-plugin::StorageEngine::~StorageEngine()
+StorageEngine::~StorageEngine()
 {
   pthread_mutex_destroy(&proto_cache_mutex);
 }
 
-void plugin::StorageEngine::setTransactionReadWrite(Session& session)
+void StorageEngine::setTransactionReadWrite(Session& session)
 {
-  Ha_trx_info *ha_info= session.getEngineInfo(this);
-
-  /*
-    When a storage engine method is called, the transaction must
-    have been started, unless it's a DDL call, for which the
-    storage engine starts the transaction internally, and commits
-    it internally, without registering in the ha_list.
-    Unfortunately here we can't know know for sure if the engine
-    has registered the transaction or not, so we must check.
-  */
-  if (ha_info->is_started())
-  {
-    /*
-     * table_share can be NULL in plugin::StorageEngine::dropTable().
-     */
-    ha_info->set_trx_read_write();
-  }
+  TransactionContext &statement_ctx= session.transaction.stmt;
+  statement_ctx.markModifiedNonTransData();
 }
 
-
-
-int plugin::StorageEngine::doRenameTable(Session *,
+int StorageEngine::doRenameTable(Session *,
                                          const char *from,
                                          const char *to)
 {
@@ -153,7 +127,7 @@
   @retval
     !0  Error
 */
-int plugin::StorageEngine::doDropTable(Session&,
+int StorageEngine::doDropTable(Session&,
                                        const string table_path)
 {
   int error= 0;
@@ -176,7 +150,7 @@
   return error;
 }
 
-const char *plugin::StorageEngine::checkLowercaseNames(const char *path,
+const char *StorageEngine::checkLowercaseNames(const char *path,
                                                        char *tmp_path)
 {
   if (flags.test(HTON_BIT_FILE_BASED))
@@ -199,14 +173,11 @@
 }
 
 
-bool plugin::StorageEngine::addPlugin(plugin::StorageEngine *engine)
+bool StorageEngine::addPlugin(StorageEngine *engine)
 {
 
   vector_of_engines.push_back(engine);
 
-  if (engine->check_flag(HTON_BIT_DOES_TRANSACTIONS))
-    vector_of_transactional_engines.push_back(engine);
-
   if (engine->getTableDefinitionFileExtension().length())
   {
     assert(engine->getTableDefinitionFileExtension().length() == DEFAULT_DEFINITION_FILE_EXT.length());
@@ -216,19 +187,18 @@
   return false;
 }
 
-void plugin::StorageEngine::removePlugin(plugin::StorageEngine *)
+void StorageEngine::removePlugin(StorageEngine *)
 {
   if (shutdown_has_begun == false)
   {
     vector_of_engines.clear();
-    vector_of_transactional_engines.clear();
 
     shutdown_has_begun= true;
   }
 }
 
 class FindEngineByName
-  : public unary_function<plugin::StorageEngine *, bool>
+  : public unary_function<StorageEngine *, bool>
 {
   const string target;
 public:
@@ -245,7 +215,7 @@
   }
 };
 
-plugin::StorageEngine *plugin::StorageEngine::findByName(string find_str)
+StorageEngine *StorageEngine::findByName(string find_str)
 {
   transform(find_str.begin(), find_str.end(),
             find_str.begin(), ::tolower);
@@ -264,7 +234,7 @@
   return NULL;
 }
 
-plugin::StorageEngine *plugin::StorageEngine::findByName(Session& session,
+StorageEngine *StorageEngine::findByName(Session& session,
                                                          string find_str)
 {
   
@@ -288,7 +258,7 @@
 }
 
 class StorageEngineCloseConnection
-: public unary_function<plugin::StorageEngine *, void>
+: public unary_function<StorageEngine *, void>
 {
   Session *session;
 public:
@@ -308,66 +278,24 @@
   @note
     don't bother to rollback here, it's done already
 */
-void plugin::StorageEngine::closeConnection(Session* session)
+void StorageEngine::closeConnection(Session* session)
 {
   for_each(vector_of_engines.begin(), vector_of_engines.end(),
            StorageEngineCloseConnection(session));
 }
 
-void plugin::StorageEngine::dropDatabase(char* path)
+void StorageEngine::dropDatabase(char* path)
 {
   for_each(vector_of_engines.begin(), vector_of_engines.end(),
-           bind2nd(mem_fun(&plugin::StorageEngine::drop_database),path));
-}
-
-int plugin::StorageEngine::commitOrRollbackByXID(XID *xid, bool commit)
-{
-  vector<int> results;
-  
-  if (commit)
-    transform(vector_of_engines.begin(), vector_of_engines.end(), results.begin(),
-              bind2nd(mem_fun(&plugin::StorageEngine::commit_by_xid),xid));
-  else
-    transform(vector_of_engines.begin(), vector_of_engines.end(), results.begin(),
-              bind2nd(mem_fun(&plugin::StorageEngine::rollback_by_xid),xid));
-
-  if (find_if(results.begin(), results.end(), bind2nd(equal_to<int>(),0))
-         == results.end())
-    return 1;
-  return 0;
-}
-
-/**
-  @details
-  This function should be called when MySQL sends rows of a SELECT result set
-  or the EOF mark to the client. It releases a possible adaptive hash index
-  S-latch held by session in InnoDB and also releases a possible InnoDB query
-  FIFO ticket to enter InnoDB. To save CPU time, InnoDB allows a session to
-  keep them over several calls of the InnoDB Cursor interface when a join
-  is executed. But when we let the control to pass to the client they have
-  to be released because if the application program uses mysql_use_result(),
-  it may deadlock on the S-latch if the application on another connection
-  performs another SQL query. In MySQL-4.1 this is even more important because
-  there a connection can have several SELECT queries open at the same time.
-
-  @param session           the thread handle of the current connection
-
-  @return
-    always 0
-*/
-int plugin::StorageEngine::releaseTemporaryLatches(Session *session)
-{
-  for_each(vector_of_transactional_engines.begin(), vector_of_transactional_engines.end(),
-           bind2nd(mem_fun(&plugin::StorageEngine::release_temporary_latches),session));
-  return 0;
-}
-
-bool plugin::StorageEngine::flushLogs(plugin::StorageEngine *engine)
+           bind2nd(mem_fun(&StorageEngine::drop_database),path));
+}
+
+bool StorageEngine::flushLogs(StorageEngine *engine)
 {
   if (engine == NULL)
   {
     if (find_if(vector_of_engines.begin(), vector_of_engines.end(),
-            mem_fun(&plugin::StorageEngine::flush_logs))
+            mem_fun(&StorageEngine::flush_logs))
           != vector_of_engines.begin())
       return true;
   }
@@ -380,169 +308,7 @@
   return false;
 }
 
-/**
-  recover() step of xa.
-
-  @note
-    there are three modes of operation:
-    - automatic recover after a crash
-    in this case commit_list != 0, tc_heuristic_recover==0
-    all xids from commit_list are committed, others are rolled back
-    - manual (heuristic) recover
-    in this case commit_list==0, tc_heuristic_recover != 0
-    DBA has explicitly specified that all prepared transactions should
-    be committed (or rolled back).
-    - no recovery (MySQL did not detect a crash)
-    in this case commit_list==0, tc_heuristic_recover == 0
-    there should be no prepared transactions in this case.
-*/
-class XARecover : unary_function<plugin::StorageEngine *, void>
-{
-  int trans_len, found_foreign_xids, found_my_xids;
-  bool result;
-  XID *trans_list;
-  HASH *commit_list;
-  bool dry_run;
-public:
-  XARecover(XID *trans_list_arg, int trans_len_arg,
-            HASH *commit_list_arg, bool dry_run_arg) 
-    : trans_len(trans_len_arg), found_foreign_xids(0), found_my_xids(0),
-      result(false),
-      trans_list(trans_list_arg), commit_list(commit_list_arg),
-      dry_run(dry_run_arg)
-  {}
-  
-  int getForeignXIDs()
-  {
-    return found_foreign_xids; 
-  }
-
-  int getMyXIDs()
-  {
-    return found_my_xids; 
-  }
-
-  result_type operator() (argument_type engine)
-  {
-  
-    int got;
-  
-    if (engine->is_enabled())
-    {
-      while ((got= engine->recover(trans_list, trans_len)) > 0 )
-      {
-        errmsg_printf(ERRMSG_LVL_INFO,
-                      _("Found %d prepared transaction(s) in %s"),
-                      got, engine->getName().c_str());
-        for (int i=0; i < got; i ++)
-        {
-          my_xid x=trans_list[i].get_my_xid();
-          if (!x) // not "mine" - that is generated by external TM
-          {
-            xid_cache_insert(trans_list+i, XA_PREPARED);
-            found_foreign_xids++;
-            continue;
-          }
-          if (dry_run)
-          {
-            found_my_xids++;
-            continue;
-          }
-          // recovery mode
-          if (commit_list ?
-              hash_search(commit_list, (unsigned char *)&x, sizeof(x)) != 0 :
-              tc_heuristic_recover == TC_HEURISTIC_RECOVER_COMMIT)
-          {
-            engine->commit_by_xid(trans_list+i);
-          }
-          else
-          {
-            engine->rollback_by_xid(trans_list+i);
-          }
-        }
-        if (got < trans_len)
-          break;
-      }
-    }
-  }
-};
-
-int plugin::StorageEngine::recover(HASH *commit_list)
-{
-  XID *trans_list= NULL;
-  int trans_len= 0;
-
-  bool dry_run= (commit_list==0 && tc_heuristic_recover==0);
-
-  /* commit_list and tc_heuristic_recover cannot be set both */
-  assert(commit_list==0 || tc_heuristic_recover==0);
-
-  /* if either is set, total_ha_2pc must be set too */
-  if (total_ha_2pc <= 1)
-    return 0;
-
-
-#ifndef WILL_BE_DELETED_LATER
-
-  /*
-    for now, only InnoDB supports 2pc. It means we can always safely
-    rollback all pending transactions, without risking inconsistent data
-  */
-
-  assert(total_ha_2pc == 2); // only InnoDB and binlog
-  tc_heuristic_recover= TC_HEURISTIC_RECOVER_ROLLBACK; // forcing ROLLBACK
-  dry_run=false;
-#endif
-  for (trans_len= MAX_XID_LIST_SIZE ;
-       trans_list==0 && trans_len > MIN_XID_LIST_SIZE; trans_len/=2)
-  {
-    trans_list=(XID *)malloc(trans_len*sizeof(XID));
-  }
-  if (!trans_list)
-  {
-    errmsg_printf(ERRMSG_LVL_ERROR, ER(ER_OUTOFMEMORY), trans_len*sizeof(XID));
-    return(1);
-  }
-
-  if (commit_list)
-    errmsg_printf(ERRMSG_LVL_INFO, _("Starting crash recovery..."));
-
-
-  XARecover recover_func(trans_list, trans_len, commit_list, dry_run);
-  for_each(vector_of_transactional_engines.begin(), vector_of_transactional_engines.end(),
-           recover_func);
-  free(trans_list);
- 
-  if (recover_func.getForeignXIDs())
-    errmsg_printf(ERRMSG_LVL_WARN,
-                  _("Found %d prepared XA transactions"),
-                  recover_func.getForeignXIDs());
-  if (dry_run && recover_func.getMyXIDs())
-  {
-    errmsg_printf(ERRMSG_LVL_ERROR,
-                  _("Found %d prepared transactions! It means that drizzled "
-                    "was not shut down properly last time and critical "
-                    "recovery information (last binlog or %s file) was "
-                    "manually deleted after a crash. You have to start "
-                    "drizzled with the --tc-heuristic-recover switch to "
-                    "commit or rollback pending transactions."),
-                    recover_func.getMyXIDs(), opt_tc_log_file);
-    return(1);
-  }
-  if (commit_list)
-    errmsg_printf(ERRMSG_LVL_INFO, _("Crash recovery finished."));
-  return(0);
-}
-
-int plugin::StorageEngine::startConsistentSnapshot(Session *session)
-{
-  for_each(vector_of_engines.begin(), vector_of_engines.end(),
-           bind2nd(mem_fun(&plugin::StorageEngine::start_consistent_snapshot),
-                   session));
-  return 0;
-}
-
-class StorageEngineGetTableDefinition: public unary_function<plugin::StorageEngine *,bool>
+class StorageEngineGetTableDefinition: public unary_function<StorageEngine *,bool>
 {
   Session& session;
   const char* path;
@@ -611,7 +377,7 @@
   to ask engine if there are any new tables that should be written to disk
   or any dropped tables that need to be removed from disk
 */
-int plugin::StorageEngine::getTableDefinition(Session& session,
+int StorageEngine::getTableDefinition(Session& session,
                                               TableIdentifier &identifier,
                                               message::Table *table_proto)
 {
@@ -620,7 +386,7 @@
                             table_proto);
 }
 
-int plugin::StorageEngine::getTableDefinition(Session& session,
+int StorageEngine::getTableDefinition(Session& session,
                                               const char* path,
                                               const char *,
                                               const char *,
@@ -629,7 +395,7 @@
 {
   int err= ENOENT;
 
-  vector<plugin::StorageEngine *>::iterator iter=
+  vector<StorageEngine *>::iterator iter=
     find_if(vector_of_engines.begin(), vector_of_engines.end(),
             StorageEngineGetTableDefinition(session, path, NULL, NULL, true, table_proto, &err));
 
@@ -693,16 +459,16 @@
 /**
    returns ENOENT if the file doesn't exists.
 */
-int plugin::StorageEngine::dropTable(Session& session,
+int StorageEngine::dropTable(Session& session,
                                      TableIdentifier &identifier,
                                      bool generate_warning)
 {
   int error= 0;
   int error_proto;
   message::Table src_proto;
-  plugin::StorageEngine* engine;
+  StorageEngine* engine;
 
-  error_proto= plugin::StorageEngine::getTableDefinition(session,
+  error_proto= StorageEngine::getTableDefinition(session,
                                                          identifier,
                                                          &src_proto);
 
@@ -713,8 +479,7 @@
     return ER_CORRUPT_TABLE_DEFINITION;
   }
 
-  engine= plugin::StorageEngine::findByName(session,
-                                            src_proto.engine().name());
+  engine= StorageEngine::findByName(session, src_proto.engine().name());
 
   if (engine)
   {
@@ -784,7 +549,7 @@
 
    @todo refactor to remove goto
 */
-int plugin::StorageEngine::createTable(Session& session,
+int StorageEngine::createTable(Session& session,
                                        TableIdentifier &identifier,
                                        bool update_create_info,
                                        message::Table& table_proto, bool proto_used)
@@ -864,7 +629,7 @@
   return(error != 0);
 }
 
-Cursor *plugin::StorageEngine::getCursor(TableShare &share, memory::Root *alloc)
+Cursor *StorageEngine::getCursor(TableShare &share, memory::Root *alloc)
 {
   assert(enabled);
   return create(share, alloc);
@@ -873,7 +638,7 @@
 /**
   TODO -> Remove this to force all engines to implement their own file. Solves the "we only looked at dfe" problem.
 */
-void plugin::StorageEngine::doGetTableNames(CachedDirectory &directory, string&, set<string>& set_of_names)
+void StorageEngine::doGetTableNames(CachedDirectory &directory, string&, set<string>& set_of_names)
 {
   CachedDirectory::Entries entries= directory.getEntries();
 
@@ -904,7 +669,7 @@
 }
 
 class AddTableName : 
-  public unary_function<plugin::StorageEngine *, void>
+  public unary_function<StorageEngine *, void>
 {
   string db;
   CachedDirectory& directory;
@@ -925,7 +690,7 @@
   }
 };
 
-void plugin::StorageEngine::getSchemaNames(set<string>& set_of_names)
+void StorageEngine::getSchemaNames(set<string>& set_of_names)
 {
   CachedDirectory directory(drizzle_data_home, CachedDirectory::DIRECTORY);
 
@@ -950,7 +715,7 @@
 /*
   Return value is "if parsed"
 */
-bool plugin::StorageEngine::getSchemaDefinition(const std::string &schema_name, message::Schema &proto)
+bool StorageEngine::getSchemaDefinition(const std::string &schema_name, message::Schema &proto)
 {
   int ret;
 
@@ -968,7 +733,7 @@
   return ret == 0 ? true : false;
 }
 
-void plugin::StorageEngine::getTableNames(const string& db, set<string>& set_of_names)
+void StorageEngine::getTableNames(const string& db, set<string>& set_of_names)
 {
   char tmp_path[FN_REFLEN];
 
@@ -994,7 +759,7 @@
 }
 
 /* This will later be converted to TableIdentifiers */
-class DropTables: public unary_function<plugin::StorageEngine *, void>
+class DropTables: public unary_function<StorageEngine *, void>
 {
   Session &session;
   set<string>& set_of_names;
@@ -1028,7 +793,7 @@
 
   Note-> Unlike MySQL, we do not, on purpose, delete files that do not match any engines. 
 */
-void plugin::StorageEngine::removeLostTemporaryTables(Session &session, const char *directory)
+void StorageEngine::removeLostTemporaryTables(Session &session, const char *directory)
 {
   CachedDirectory dir(directory, set_of_table_definition_ext);
   set<string> set_of_table_names;
@@ -1071,7 +836,7 @@
   */
   set<string> all_exts= set_of_table_definition_ext;
 
-  for (vector<plugin::StorageEngine *>::iterator iter= vector_of_engines.begin();
+  for (vector<StorageEngine *>::iterator iter= vector_of_engines.begin();
        iter != vector_of_engines.end() ; iter++)
   {
     for (const char **ext= (*iter)->bas_ext(); *ext ; ext++)
@@ -1105,12 +870,12 @@
     - table->s->path
     - table->alias
 */
-void plugin::StorageEngine::print_error(int error, myf errflag, Table &table)
+void StorageEngine::print_error(int error, myf errflag, Table &table)
 {
   print_error(error, errflag, &table);
 }
 
-void plugin::StorageEngine::print_error(int error, myf errflag, Table *table)
+void StorageEngine::print_error(int error, myf errflag, Table *table)
 {
   int textno= ER_GET_ERRNO;
   switch (error) {
@@ -1322,13 +1087,13 @@
   @return
     Returns true if this is a temporary error
 */
-bool plugin::StorageEngine::get_error_message(int , String* )
+bool StorageEngine::get_error_message(int , String* )
 {
   return false;
 }
 
 
-void plugin::StorageEngine::print_keydup_error(uint32_t key_nr, const char *msg, Table &table)
+void StorageEngine::print_keydup_error(uint32_t key_nr, const char *msg, Table &table)
 {
   /* Write the duplicated key in the error message */
   char key[MAX_KEY_LENGTH];
@@ -1356,7 +1121,7 @@
 }
 
 
-int plugin::StorageEngine::deleteDefinitionFromPath(TableIdentifier &identifier)
+int StorageEngine::deleteDefinitionFromPath(TableIdentifier &identifier)
 {
   string path(identifier.getPath());
 
@@ -1365,7 +1130,7 @@
   return internal::my_delete(path.c_str(), MYF(0));
 }
 
-int plugin::StorageEngine::renameDefinitionFromPath(TableIdentifier &dest, TableIdentifier &src)
+int StorageEngine::renameDefinitionFromPath(TableIdentifier &dest, TableIdentifier &src)
 {
   string src_path(src.getPath());
   string dest_path(dest.getPath());
@@ -1376,7 +1141,7 @@
   return internal::my_rename(src_path.c_str(), dest_path.c_str(), MYF(MY_WME));
 }
 
-int plugin::StorageEngine::writeDefinitionFromPath(TableIdentifier &identifier, message::Table &table_proto)
+int StorageEngine::writeDefinitionFromPath(TableIdentifier &identifier, message::Table &table_proto)
 {
   string file_name(identifier.getPath());
 
@@ -1402,5 +1167,5 @@
   return 0;
 }
 
-
+} /* namespace plugin */
 } /* namespace drizzled */

=== modified file 'drizzled/plugin/storage_engine.h'
--- drizzled/plugin/storage_engine.h	2010-02-05 08:11:15 +0000
+++ drizzled/plugin/storage_engine.h	2010-02-14 20:26:43 +0000
@@ -31,6 +31,8 @@
 #include "drizzled/table_identifier.h"
 #include "drizzled/cached_directory.h"
 
+#include "drizzled/hash.h"
+
 #include <bitset>
 #include <string>
 #include <vector>
@@ -41,7 +43,6 @@
 
 class TableList;
 class Session;
-class XID;
 class Cursor;
 typedef struct st_hash HASH;
 
@@ -119,6 +120,9 @@
 namespace plugin
 {
 
+typedef hash_map<std::string, StorageEngine *> EngineMap;
+typedef std::vector<StorageEngine *> EngineVector;
+
 extern const std::string UNKNOWN_STRING;
 extern const std::string DEFAULT_DEFINITION_FILE_EXT;
 
@@ -130,8 +134,6 @@
   usually StorageEngine instance is defined statically in ha_xxx.cc as
 
   static StorageEngine { ... } xxx_engine;
-
-  savepoint_*, prepare, recover, and *_by_xid pointers can be 0.
 */
 class StorageEngine : public Plugin
 {
@@ -139,16 +141,11 @@
   typedef uint64_t Table_flags;
 
 private:
-
-  /*
-    Name used for storage engine.
-  */
-  const bool two_phase_commit;
   bool enabled;
 
   const std::bitset<HTON_BIT_SIZE> flags; /* global Cursor flags */
 
-  void setTransactionReadWrite(Session& session);
+  virtual void setTransactionReadWrite(Session& session);
 
 protected:
   std::string table_definition_ext;
@@ -183,21 +180,10 @@
   ProtoCache proto_cache;
   pthread_mutex_t proto_cache_mutex;
 
-  /**
-   * Implementing classes should override these to provide savepoint
-   * functionality.
-   */
-  virtual int savepoint_set_hook(Session *, NamedSavepoint &) { return 0; }
-
-  virtual int savepoint_rollback_hook(Session *, NamedSavepoint &) { return 0; }
-
-  virtual int savepoint_release_hook(Session *, NamedSavepoint &) { return 0; }
-
 public:
 
   StorageEngine(const std::string name_arg,
-                const std::bitset<HTON_BIT_SIZE> &flags_arg= HTON_NO_FLAGS,
-                bool support_2pc= false);
+                const std::bitset<HTON_BIT_SIZE> &flags_arg= HTON_NO_FLAGS);
 
   virtual ~StorageEngine();
 
@@ -241,11 +227,6 @@
   inline uint32_t getSlot (void) const { return slot; }
   inline void setSlot (uint32_t value) { slot= value; }
 
-  bool has_2pc()
-  {
-    return two_phase_commit;
-  }
-
 
   bool is_enabled() const
   {
@@ -281,50 +262,6 @@
   {
     return 0;
   }
-  /*
-    'all' is true if it's a real commit, that makes persistent changes
-    'all' is false if it's not in fact a commit but an end of the
-    statement that is part of the transaction.
-    NOTE 'all' is also false in auto-commit mode where 'end of statement'
-    and 'real commit' mean the same event.
-  */
-  virtual int  commit(Session *, bool)
-  {
-    return 0;
-  }
-
-  virtual int  rollback(Session *, bool)
-  {
-    return 0;
-  }
-
-  /*
-    The void * points to an uninitialized storage area of requested size
-    (see savepoint_offset description)
-  */
-  int savepoint_set(Session *session, NamedSavepoint &sp)
-  {
-    return savepoint_set_hook(session, sp);
-  }
-
-  /*
-    The void * points to a storage area, that was earlier passed
-    to the savepoint_set call
-  */
-  int savepoint_rollback(Session *session, NamedSavepoint &sp)
-  {
-     return savepoint_rollback_hook(session, sp);
-  }
-
-  int savepoint_release(Session *session, NamedSavepoint &sp)
-  {
-    return savepoint_release_hook(session, sp);
-  }
-
-  virtual int  prepare(Session *, bool) { return 0; }
-  virtual int  recover(XID *, uint32_t) { return 0; }
-  virtual int  commit_by_xid(XID *) { return 0; }
-  virtual int  rollback_by_xid(XID *) { return 0; }
   virtual Cursor *create(TableShare &, memory::Root *)= 0;
   /* args: path */
   virtual void drop_database(char*) { }
@@ -335,8 +272,6 @@
     return false;
   }
 
-  virtual int release_temporary_latches(Session *) { return false; }
-
   /**
     If frm_error() is called then we will use this to find out what file
     extentions exist for the storage engine. This is also used by the default
@@ -397,11 +332,7 @@
                                            std::string find_str);
   static void closeConnection(Session* session);
   static void dropDatabase(char* path);
-  static int commitOrRollbackByXID(XID *xid, bool commit);
-  static int releaseTemporaryLatches(Session *session);
   static bool flushLogs(plugin::StorageEngine *db_type);
-  static int recover(HASH *commit_list);
-  static int startConsistentSnapshot(Session *session);
   static int dropTable(Session& session,
                        TableIdentifier &identifier,
                        bool generate_warning);

=== added file 'drizzled/plugin/transactional_storage_engine.cc'
--- drizzled/plugin/transactional_storage_engine.cc	1970-01-01 00:00:00 +0000
+++ drizzled/plugin/transactional_storage_engine.cc	2010-02-14 19:27:57 +0000
@@ -0,0 +1,116 @@
+/* -*- mode: c++; c-basic-offset: 2; indent-tabs-mode: nil; -*-
+ *  vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
+ *
+ *  Copyright (C) 2008 Sun Microsystems
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; version 2 of the License.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ */
+
+#include "config.h"
+
+#include "drizzled/plugin/transactional_storage_engine.h"
+#include "drizzled/resource_context.h"
+#include "drizzled/session.h"
+
+#include <vector>
+#include <algorithm>
+#include <functional>
+
+using namespace std;
+
+namespace drizzled
+{
+
+namespace plugin
+{
+
+static vector<TransactionalStorageEngine *> vector_of_transactional_engines;
+
+TransactionalStorageEngine::TransactionalStorageEngine(const string name_arg,
+                                     const bitset<HTON_BIT_SIZE> &flags_arg,
+                                     bool two_phase_commit_arg)
+    : StorageEngine(name_arg, flags_arg),
+      two_phase_commit(two_phase_commit_arg)
+{
+}
+
+TransactionalStorageEngine::~TransactionalStorageEngine()
+{
+}
+
+void TransactionalStorageEngine::setTransactionReadWrite(Session& session)
+{
+  ResourceContext *resource_context= session.getResourceContext(this);
+
+  /*
+    When a storage engine method is called, the transaction must
+    have been started, unless it's a DDL call, for which the
+    storage engine starts the transaction internally, and commits
+    it internally, without registering in the ha_list.
+    Unfortunately here we can't know know for sure if the engine
+    has registered the transaction or not, so we must check.
+  */
+  if (resource_context->isStarted())
+  {
+    resource_context->markModifiedData();
+  }
+}
+
+/**
+  @details
+  This function should be called when MySQL sends rows of a SELECT result set
+  or the EOF mark to the client. It releases a possible adaptive hash index
+  S-latch held by session in InnoDB and also releases a possible InnoDB query
+  FIFO ticket to enter InnoDB. To save CPU time, InnoDB allows a session to
+  keep them over several calls of the InnoDB Cursor interface when a join
+  is executed. But when we let the control to pass to the client they have
+  to be released because if the application program uses mysql_use_result(),
+  it may deadlock on the S-latch if the application on another connection
+  performs another SQL query. In MySQL-4.1 this is even more important because
+  there a connection can have several SELECT queries open at the same time.
+
+  @param session           the thread handle of the current connection
+
+  @return
+    always 0
+*/
+int TransactionalStorageEngine::releaseTemporaryLatches(Session *session)
+{
+  for_each(vector_of_transactional_engines.begin(), vector_of_transactional_engines.end(),
+           bind2nd(mem_fun(&TransactionalStorageEngine::doReleaseTemporaryLatches),session));
+  return 0;
+}
+
+int TransactionalStorageEngine::startConsistentSnapshot(Session *session)
+{
+  for_each(vector_of_transactional_engines.begin(), vector_of_transactional_engines.end(),
+           bind2nd(mem_fun(&TransactionalStorageEngine::doStartConsistentSnapshot),
+                   session));
+  return 0;
+}
+
+bool TransactionalStorageEngine::addPlugin(TransactionalStorageEngine *engine)
+{
+  vector_of_transactional_engines.push_back(engine);
+
+  return StorageEngine::addPlugin(engine);
+}
+
+void TransactionalStorageEngine::removePlugin(TransactionalStorageEngine *)
+{
+  vector_of_transactional_engines.clear();
+}
+
+} /* namespace plugin */
+} /* namespace drizzled */

=== added file 'drizzled/plugin/transactional_storage_engine.h'
--- drizzled/plugin/transactional_storage_engine.h	1970-01-01 00:00:00 +0000
+++ drizzled/plugin/transactional_storage_engine.h	2010-02-14 19:27:57 +0000
@@ -0,0 +1,164 @@
+/* -*- mode: c++; c-basic-offset: 2; indent-tabs-mode: nil; -*-
+ *  vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
+ *
+ *  Copyright (C) 2008 Sun Microsystems
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; version 2 of the License.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ */
+
+#ifndef DRIZZLED_PLUGIN_TRANSACTIONAL_STORAGE_ENGINE_H
+#define DRIZZLED_PLUGIN_TRANSACTIONAL_STORAGE_ENGINE_H
+
+#include "drizzled/plugin/storage_engine.h"
+
+namespace drizzled
+{
+
+namespace plugin
+{
+
+/**
+ * A type of storage engine which supports SQL transactions.
+ *
+ * This class adds the SQL transactional API to the regular
+ * storage engine.  In other words, it adds support for the
+ * following SQL statements:
+ *
+ * START TRANSACTION;
+ * COMMIT;
+ * ROLLBACK;
+ * ROLLBACK TO SAVEPOINT;
+ * SET SAVEPOINT;
+ * RELEASE SAVEPOINT;
+ *
+ * @note
+ *
+ * This class does not implement the XA protocol (two phase commit).
+ * There is an XaStorageEngine class which extends this class that
+ * exposes the XA API.
+ *
+ * @todo
+ *
+ * kill two_phase_commit member. Use an HTON flag if
+ * absolutely needed to keep.
+ */
+class TransactionalStorageEngine :public StorageEngine
+{
+public:
+  TransactionalStorageEngine(const std::string name_arg,
+                             const std::bitset<HTON_BIT_SIZE> &flags_arg= HTON_NO_FLAGS,
+                             bool two_phase_commit= false);
+
+  virtual ~TransactionalStorageEngine();
+
+  int commit(Session *session, bool normal_transaction)
+  {
+    return doCommit(session, normal_transaction);
+  }
+
+  int rollback(Session *session, bool normal_transaction)
+  {
+    return doRollback(session, normal_transaction);
+  }
+
+  int setSavepoint(Session *session, NamedSavepoint &sp)
+  {
+    return doSetSavepoint(session, sp);
+  }
+
+  int rollbackToSavepoint(Session *session, NamedSavepoint &sp)
+  {
+     return doRollbackToSavepoint(session, sp);
+  }
+
+  int releaseSavepoint(Session *session, NamedSavepoint &sp)
+  {
+    return doReleaseSavepoint(session, sp);
+  }
+
+  bool hasTwoPhaseCommit()
+  {
+    return two_phase_commit;
+  }
+
+  /** 
+   * The below static class methods wrap the interaction
+   * of the vector of transactional storage engines.
+   *
+   * @todo kill these. they belong in TransactionServices.
+   */
+  /**
+   * @todo Kill this one entirely.  It's implementation, not interface...
+   */
+  static int releaseTemporaryLatches(Session *session);
+  static int startConsistentSnapshot(Session *session);
+
+  /* Class Methods for operating on plugin */
+  static bool addPlugin(plugin::TransactionalStorageEngine *engine);
+  static void removePlugin(plugin::TransactionalStorageEngine *engine);
+
+private:
+  void setTransactionReadWrite(Session& session);
+  /**
+   * Implementing classes should override these to provide savepoint
+   * functionality.
+   */
+  virtual int doSetSavepoint(Session *session, NamedSavepoint &savepoint)= 0;
+  virtual int doRollbackToSavepoint(Session *session, NamedSavepoint &savepoint)= 0;
+  virtual int doReleaseSavepoint(Session *session, NamedSavepoint &savepoint)= 0;
+  
+  /**
+   * Commits either the "statement transaction" or the "normal transaction".
+   *
+   * @param[in] The Session
+   * @param[in] true if it's a real commit, that makes persistent changes
+   *            false if it's not in fact a commit but an end of the
+   *            statement that is part of the transaction.
+   * @note
+   *
+   * 'normal_transaction' is also false in auto-commit mode where 'end of statement'
+   * and 'real commit' mean the same event.
+   */
+  virtual int doCommit(Session *session, bool normal_transaction)= 0;
+
+  /**
+   * Rolls back either the "statement transaction" or the "normal transaction".
+   *
+   * @param[in] The Session
+   * @param[in] true if it's a real commit, that makes persistent changes
+   *            false if it's not in fact a commit but an end of the
+   *            statement that is part of the transaction.
+   * @note
+   *
+   * 'normal_transaction' is also false in auto-commit mode where 'end of statement'
+   * and 'real commit' mean the same event.
+   */
+  virtual int doRollback(Session *session, bool normal_transaction)= 0;
+  virtual int doReleaseTemporaryLatches(Session *session)
+  {
+    (void) session;
+    return 0;
+  }
+  virtual int doStartConsistentSnapshot(Session *session)
+  {
+    (void) session;
+    return 0;
+  }
+  const bool two_phase_commit;
+};
+
+} /* namespace plugin */
+} /* namespace drizzled */
+
+#endif /* DRIZZLED_PLUGIN_TRANSACTIONAL_STORAGE_ENGINE_H */

=== added file 'drizzled/plugin/xa_storage_engine.cc'
--- drizzled/plugin/xa_storage_engine.cc	1970-01-01 00:00:00 +0000
+++ drizzled/plugin/xa_storage_engine.cc	2010-02-14 19:27:57 +0000
@@ -0,0 +1,237 @@
+/* -*- mode: c++; c-basic-offset: 2; indent-tabs-mode: nil; -*-
+ *  vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
+ *
+ *  Copyright (C) 2008 Sun Microsystems
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; version 2 of the License.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ */
+
+#include "config.h"
+
+#include "drizzled/my_hash.h"
+#include "drizzled/cached_directory.h"
+
+#include <drizzled/definitions.h>
+#include <drizzled/session.h>
+#include <drizzled/error.h>
+#include <drizzled/gettext.h>
+#include <drizzled/plugin/xa_storage_engine.h>
+#include "drizzled/xid.h"
+
+#include "drizzled/hash.h"
+
+#include <string>
+#include <vector>
+#include <algorithm>
+#include <functional>
+
+using namespace std;
+
+namespace drizzled
+{
+
+namespace plugin
+{
+
+static vector<XaStorageEngine *> vector_of_xa_engines;
+
+XaStorageEngine::XaStorageEngine(const string name_arg,
+                                 const bitset<HTON_BIT_SIZE> &flags_arg)
+    : TransactionalStorageEngine(name_arg, flags_arg, true)
+{}
+
+XaStorageEngine::~XaStorageEngine()
+{}
+
+int XaStorageEngine::commitOrRollbackXID(XID *xid, bool commit)
+{
+  vector<int> results;
+  
+  if (commit)
+    transform(vector_of_xa_engines.begin(), vector_of_xa_engines.end(), results.begin(),
+              bind2nd(mem_fun(&XaStorageEngine::commitXid), xid));
+  else
+    transform(vector_of_xa_engines.begin(), vector_of_xa_engines.end(), results.begin(),
+              bind2nd(mem_fun(&XaStorageEngine::rollbackXid), xid));
+
+  if (find_if(results.begin(), results.end(), bind2nd(equal_to<int>(),0))
+         == results.end())
+    return 1;
+  return 0;
+}
+
+/**
+  recover() step of xa.
+
+  @note
+    there are three modes of operation:
+    - automatic recover after a crash
+    in this case commit_list != 0, tc_heuristic_recover==0
+    all xids from commit_list are committed, others are rolled back
+    - manual (heuristic) recover
+    in this case commit_list==0, tc_heuristic_recover != 0
+    DBA has explicitly specified that all prepared transactions should
+    be committed (or rolled back).
+    - no recovery (MySQL did not detect a crash)
+    in this case commit_list==0, tc_heuristic_recover == 0
+    there should be no prepared transactions in this case.
+*/
+class XaRecover : unary_function<XaStorageEngine *, void>
+{
+  int trans_len, found_foreign_xids, found_my_xids;
+  bool result;
+  XID *trans_list;
+  HASH *commit_list;
+  bool dry_run;
+public:
+  XaRecover(XID *trans_list_arg, int trans_len_arg,
+            HASH *commit_list_arg, bool dry_run_arg) 
+    : trans_len(trans_len_arg), found_foreign_xids(0), found_my_xids(0),
+      result(false),
+      trans_list(trans_list_arg), commit_list(commit_list_arg),
+      dry_run(dry_run_arg)
+  {}
+  
+  int getForeignXIDs()
+  {
+    return found_foreign_xids; 
+  }
+
+  int getMyXIDs()
+  {
+    return found_my_xids; 
+  }
+
+  result_type operator() (argument_type engine)
+  {
+  
+    int got;
+  
+    if (engine->is_enabled())
+    {
+      while ((got= engine->recover(trans_list, trans_len)) > 0 )
+      {
+        errmsg_printf(ERRMSG_LVL_INFO,
+                      _("Found %d prepared transaction(s) in %s"),
+                      got, engine->getName().c_str());
+        for (int i=0; i < got; i ++)
+        {
+          my_xid x=trans_list[i].get_my_xid();
+          if (!x) // not "mine" - that is generated by external TM
+          {
+            xid_cache_insert(trans_list+i, XA_PREPARED);
+            found_foreign_xids++;
+            continue;
+          }
+          if (dry_run)
+          {
+            found_my_xids++;
+            continue;
+          }
+          // recovery mode
+          if (commit_list ?
+              hash_search(commit_list, (unsigned char *)&x, sizeof(x)) != 0 :
+              tc_heuristic_recover == TC_HEURISTIC_RECOVER_COMMIT)
+          {
+            engine->commitXid(trans_list+i);
+          }
+          else
+          {
+            engine->rollbackXid(trans_list+i);
+          }
+        }
+        if (got < trans_len)
+          break;
+      }
+    }
+  }
+};
+
+int XaStorageEngine::recoverAllXids(HASH *commit_list)
+{
+  XID *trans_list= NULL;
+  int trans_len= 0;
+
+  bool dry_run= (commit_list==0 && tc_heuristic_recover==0);
+
+  /* commit_list and tc_heuristic_recover cannot be set both */
+  assert(commit_list==0 || tc_heuristic_recover==0);
+
+  /* if either is set, total_ha_2pc must be set too */
+  if (total_ha_2pc <= 1)
+    return 0;
+
+  /*
+    for now, only InnoDB supports 2pc. It means we can always safely
+    rollback all pending transactions, without risking inconsistent data
+  */
+
+  assert(total_ha_2pc == 2); // only InnoDB and binlog
+  tc_heuristic_recover= TC_HEURISTIC_RECOVER_ROLLBACK; // forcing ROLLBACK
+  dry_run=false;
+  for (trans_len= MAX_XID_LIST_SIZE ;
+       trans_list==0 && trans_len > MIN_XID_LIST_SIZE; trans_len/=2)
+  {
+    trans_list=(XID *)malloc(trans_len*sizeof(XID));
+  }
+  if (!trans_list)
+  {
+    errmsg_printf(ERRMSG_LVL_ERROR, ER(ER_OUTOFMEMORY), trans_len*sizeof(XID));
+    return(1);
+  }
+
+  if (commit_list)
+    errmsg_printf(ERRMSG_LVL_INFO, _("Starting crash recovery..."));
+
+  XaRecover recover_func(trans_list, trans_len, commit_list, dry_run);
+  for_each(vector_of_xa_engines.begin(),
+           vector_of_xa_engines.end(),
+           recover_func);
+  free(trans_list);
+ 
+  if (recover_func.getForeignXIDs())
+    errmsg_printf(ERRMSG_LVL_WARN,
+                  _("Found %d prepared XA transactions"),
+                  recover_func.getForeignXIDs());
+  if (dry_run && recover_func.getMyXIDs())
+  {
+    errmsg_printf(ERRMSG_LVL_ERROR,
+                  _("Found %d prepared transactions! It means that drizzled "
+                    "was not shut down properly last time and critical "
+                    "recovery information (last binlog or %s file) was "
+                    "manually deleted after a crash. You have to start "
+                    "drizzled with the --tc-heuristic-recover switch to "
+                    "commit or rollback pending transactions."),
+                    recover_func.getMyXIDs(), opt_tc_log_file);
+    return(1);
+  }
+  if (commit_list)
+    errmsg_printf(ERRMSG_LVL_INFO, _("Crash recovery finished."));
+  return(0);
+}
+
+bool XaStorageEngine::addPlugin(XaStorageEngine *engine)
+{
+  vector_of_xa_engines.push_back(engine);
+
+  return TransactionalStorageEngine::addPlugin(engine);
+}
+
+void XaStorageEngine::removePlugin(XaStorageEngine *)
+{
+  vector_of_xa_engines.clear();
+}
+
+} /* namespace plugin */
+} /* namespace drizzled */

=== added file 'drizzled/plugin/xa_storage_engine.h'
--- drizzled/plugin/xa_storage_engine.h	1970-01-01 00:00:00 +0000
+++ drizzled/plugin/xa_storage_engine.h	2010-02-14 19:27:57 +0000
@@ -0,0 +1,108 @@
+/* -*- mode: c++; c-basic-offset: 2; indent-tabs-mode: nil; -*-
+ *  vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
+ *
+ *  Copyright (C) 2008 Sun Microsystems
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; version 2 of the License.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ */
+
+#ifndef DRIZZLED_PLUGIN_XA_STORAGE_ENGINE_H
+#define DRIZZLED_PLUGIN_XA_STORAGE_ENGINE_H
+
+#include "drizzled/plugin/transactional_storage_engine.h"
+
+namespace drizzled
+{
+
+class XID;
+
+namespace plugin
+{
+
+/**
+ * A type of storage engine which supports distributed
+ * transactions in the XA protocol.
+ */
+class XaStorageEngine :public TransactionalStorageEngine
+{
+public:
+  XaStorageEngine(const std::string name_arg,
+                  const std::bitset<HTON_BIT_SIZE> &flags_arg= HTON_NO_FLAGS);
+
+  virtual ~XaStorageEngine();
+
+  int prepare(Session *session, bool normal_transaction)
+  {
+    return doPrepare(session, normal_transaction);
+  }
+
+  int commitXid(XID *xid)
+  {
+    return doCommitXid(xid);
+  }
+
+  int rollbackXid(XID *xid)
+  {
+    return doRollbackXid(xid);
+  }
+
+  int recover(XID * append_to, size_t len)
+  {
+    return doRecover(append_to, len);
+  }
+
+  /** 
+   * The below static class methods wrap the interaction
+   * of the vector of registered XA storage engines.
+   */
+  static int commitOrRollbackXID(XID *xid, bool commit);
+  static int recoverAllXids(HASH *commit_list);
+
+  /* Class Methods for operating on plugin */
+  static bool addPlugin(plugin::XaStorageEngine *engine);
+  static void removePlugin(plugin::XaStorageEngine *engine);
+
+private:
+  /**
+   * Does the PREPARE stage of the two-phase commit.
+   */
+  virtual int doPrepare(Session *session, bool normal_transaction)= 0;
+  /**
+   * Rolls back a transaction identified by a XID.
+   */
+  virtual int doRollbackXid(XID *xid)= 0;
+  /**
+   * Commits a transaction identified by a XID.
+   */
+  virtual int doCommitXid(XID *xid)= 0;
+  /**
+   * Notifies the transaction manager of any transactions
+   * which had been marked prepared but not committed at
+   * crash time or that have been heurtistically completed
+   * by the storage engine.
+   *
+   * @param[out] Reference to a vector of XIDs to add to
+   *
+   * @retval
+   *  Returns the number of transactions left to recover
+   *  for this engine.
+   */
+  virtual int doRecover(XID * append_to, size_t len)= 0;
+
+};
+
+} /* namespace plugin */
+} /* namespace drizzled */
+
+#endif /* DRIZZLED_PLUGIN_XA_STORAGE_ENGINE_H */

=== renamed file 'drizzled/ha_trx_info.cc' => 'drizzled/resource_context.cc'
--- drizzled/ha_trx_info.cc	2010-02-04 08:14:46 +0000
+++ drizzled/resource_context.cc	2010-02-06 02:38:18 +0000
@@ -18,81 +18,49 @@
  */
 
 #include "config.h"
-#include <drizzled/ha_trx_info.h>
-#include <drizzled/plugin/storage_engine.h>
-#include <drizzled/session.h>
+#include "drizzled/resource_context.h"
+
+#include <cassert>
+
+using namespace std;
 
 namespace drizzled
 {
 
-void Ha_trx_info::register_ha(Session_TRANS *trans,
-                              plugin::StorageEngine *engine_arg)
-{
-  assert(m_flags == 0);
-  assert(m_engine == NULL);
-  assert(m_next == NULL);
-
-  m_engine= engine_arg;
-  m_flags= (int) TRX_READ_ONLY; /* Assume read-only at start. */
-
-  m_next= trans->ha_list;
-  trans->ha_list= this;
-}
-
-
 /** Clear, prepare for reuse. */
-void Ha_trx_info::reset()
-{
-  m_next= NULL;
-  m_engine= NULL;
-  m_flags= 0;
-}
-
-void Ha_trx_info::set_trx_read_write()
-{
-  assert(is_started());
-  m_flags|= (int) TRX_READ_WRITE;
-}
-
-
-bool Ha_trx_info::is_trx_read_write() const
-{
-  assert(is_started());
-  return m_flags & (int) TRX_READ_WRITE;
-}
-
-
-bool Ha_trx_info::is_started() const
-{
-  return m_engine != NULL;
-}
-
-
-/** Mark this transaction read-write if the argument is read-write. */
-void Ha_trx_info::coalesce_trx_with(const Ha_trx_info *stmt_trx)
+void ResourceContext::reset()
+{
+  resource= NULL;
+  modified_data= false;
+}
+
+void ResourceContext::markModifiedData()
+{
+  assert(isStarted());
+  modified_data= true;
+}
+
+bool ResourceContext::hasModifiedData() const
+{
+  assert(isStarted());
+  return modified_data;
+}
+
+bool ResourceContext::isStarted() const
+{
+  return resource != NULL;
+}
+
+void ResourceContext::coalesceWith(const ResourceContext *stmt_ctx)
 {
   /*
     Must be called only after the transaction has been started.
     Can be called many times, e.g. when we have many
     read-write statements in a transaction.
   */
-  assert(is_started());
-  if (stmt_trx->is_trx_read_write())
-    set_trx_read_write();
-}
-
-
-Ha_trx_info *Ha_trx_info::next() const
-{
-  assert(is_started());
-  return m_next;
-}
-
-
-plugin::StorageEngine *Ha_trx_info::engine() const
-{
-  assert(is_started());
-  return m_engine;
+  assert(isStarted());
+  if (stmt_ctx->hasModifiedData())
+    markModifiedData();
 }
 
 } /* namespace drizzled */

=== renamed file 'drizzled/ha_trx_info.h' => 'drizzled/resource_context.h'
--- drizzled/ha_trx_info.h	2010-02-04 08:14:46 +0000
+++ drizzled/resource_context.h	2010-02-06 02:38:18 +0000
@@ -17,75 +17,99 @@
  *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#ifndef DRIZZLED_HA_TRX_INFO_H
-#define DRIZZLED_HA_TRX_INFO_H
+#ifndef DRIZZLED_RESOURCE_CONTEXT_H
+#define DRIZZLED_RESOURCE_CONTEXT_H
 
+#include <cstddef>
 
 namespace drizzled
 {
 
-class Session_TRANS;
 namespace plugin
 {
 class StorageEngine;
 }
 
-
 /**
-  Either statement transaction or normal transaction - related
-  thread-specific storage engine data.
-
-  If a storage engine participates in a statement/transaction,
-  an instance of this class is present in
-  session->transaction.{stmt|all}.ha_list. The addition to
-  {stmt|all}.ha_list is made by trans_register_ha().
-
-  When it's time to commit or rollback, each element of ha_list
-  is used to access storage engine's prepare()/commit()/rollback()
-  methods, and also to evaluate if a full two phase commit is
-  necessary.
-
-  @sa General description of transaction handling in handler.cc.
-*/
-
-class Ha_trx_info
+ * Either statement transaction or normal transaction - related
+ * session-specific storage engine data.
+ *
+ * If a storage engine participates in a statement/transaction,
+ * an instance of this class is present in
+ * session->transaction.{stmt|all}.resource_contexts.
+ *
+ * When it's time to commit or rollback, each element of ha_list
+ * is used to access resource manager's prepare()/commit()/rollback()
+ * methods, and also to evaluate if a full two phase commit is
+ * necessary.
+ * 
+ * @sa General description of transaction handling in drizzled/transaction_services.cc.
+ */
+class ResourceContext
 {
 public:
-  /** Register this storage engine in the given transaction context. */
-  void register_ha(Session_TRANS *trans,
-                   plugin::StorageEngine *engine_arg);
+  ResourceContext() :
+    resource(NULL),
+    modified_data(false)
+  {}
 
   /** Clear, prepare for reuse. */
   void reset();
-  Ha_trx_info() { reset(); }
-
-  void set_trx_read_write();
-  bool is_trx_read_write() const;
-  bool is_started() const;
-
-  /** Mark this transaction read-write if the argument is read-write. */
-  void coalesce_trx_with(const Ha_trx_info *stmt_trx);
-  Ha_trx_info *next() const;
-  plugin::StorageEngine *engine() const;
-
+
+  /**
+   * Marks that the underlying resource manager
+   * has modified data state.
+   */
+  void markModifiedData();
+
+  /**
+   * Returns true if the underlying resource manager
+   * has modified data state.
+   */
+  bool hasModifiedData() const;
+
+  /**
+   * Returns true if the underlying resource
+   * manager has registered with the transaction
+   * manager for this transaction.
+   */
+  bool isStarted() const;
+
+  /** 
+   * Mark this context as modifying data if the argument has also modified data
+   */
+  void coalesceWith(const ResourceContext *stmt_trx);
+
+  /**
+   * Returns the underlying resource manager
+   * this context tracks.
+   */
+  drizzled::plugin::StorageEngine *getResource() const
+  {
+    return resource;
+  }
+
+  /**
+   * Sets the underlying resource
+   */
+  void setResource(drizzled::plugin::StorageEngine *in_engine)
+  {
+    resource= in_engine;
+  }
 private:
-  enum { TRX_READ_ONLY= 0, TRX_READ_WRITE= 1 };
-  /** Auxiliary, used for ha_list management */
-  Ha_trx_info *m_next;
-  /**
-    Although a given Ha_trx_info instance is currently always used
-    for the same storage engine, 'engine' is not-NULL only when the
-    corresponding storage is a part of a transaction.
-  */
-  plugin::StorageEngine *m_engine;
-  /**
-    Transaction flags related to this engine.
-    Not-null only if this instance is a part of transaction.
-    May assume a combination of enum values above.
-  */
-  unsigned char       m_flags;
+  /**
+    Although a given ResourceContext instance is always used
+    for the same resource manager, 'resource' is not-NULL only when the
+    corresponding resource manager is a part of a transaction.
+  */
+  drizzled::plugin::StorageEngine *resource;
+  /**
+   * Whether the underlying resource manager has changed
+   * some data state.
+   */
+  bool modified_data;
 };
 
 } /* namespace drizzled */
 
-#endif /* DRIZZLED_HA_TRX_INFO_H */
+#endif /* DRIZZLED_RESOURCE_CONTEXT_H */

=== modified file 'drizzled/select_send.h'
--- drizzled/select_send.h	2010-02-04 08:14:46 +0000
+++ drizzled/select_send.h	2010-02-14 19:27:57 +0000
@@ -22,6 +22,7 @@
 #define DRIZZLED_SELECT_SEND_H
 
 #include <drizzled/plugin/client.h>
+#include <drizzled/plugin/transactional_storage_engine.h>
 
 namespace drizzled
 {
@@ -42,7 +43,7 @@
       InnoDB adaptive hash S-latch to avoid thread deadlocks if it was reserved
       by session
     */
-    plugin::StorageEngine::releaseTemporaryLatches(session);
+    plugin::TransactionalStorageEngine::releaseTemporaryLatches(session);
 
     /* Unlock tables before sending packet to gain some speed */
     if (session->lock)
@@ -95,7 +96,7 @@
       InnoDB adaptive hash S-latch to avoid thread deadlocks if it was reserved
       by session
     */
-    plugin::StorageEngine::releaseTemporaryLatches(session);
+    plugin::TransactionalStorageEngine::releaseTemporaryLatches(session);
 
     List_iterator_fast<Item> li(items);
     char buff[MAX_FIELD_WIDTH];

=== modified file 'drizzled/session.cc'
--- drizzled/session.cc	2010-02-18 00:46:18 +0000
+++ drizzled/session.cc	2010-02-19 16:11:03 +0000
@@ -39,6 +39,7 @@
 #include <drizzled/plugin/client.h>
 #include "drizzled/plugin/scheduler.h"
 #include "drizzled/plugin/authentication.h"
+#include "drizzled/plugin/transactional_storage_engine.h"
 #include "drizzled/probes.h"
 #include "drizzled/table_proto.h"
 #include "drizzled/db.h"
@@ -148,10 +149,10 @@
   return static_cast<void **>(&ha_data[engine->slot].ha_ptr);
 }
 
-Ha_trx_info *Session::getEngineInfo(const plugin::StorageEngine *engine,
-                                    size_t index)
+ResourceContext *Session::getResourceContext(const plugin::StorageEngine *engine,
+                                             size_t index)
 {
-  return &ha_data[engine->getSlot()].ha_info[index];
+  return &ha_data[engine->getSlot()].resource_context[index];
 }
 
 extern "C"
@@ -256,7 +257,6 @@
   else
     options &= ~OPTION_BIG_SELECTS;
 
-  transaction.all.modified_non_trans_table= transaction.stmt.modified_non_trans_table= false;
   open_options=ha_open_options;
   update_lock_default= TL_WRITE;
   session_tx_isolation= (enum_tx_isolation) variables.tx_isolation;
@@ -790,7 +790,6 @@
       if (transaction_services.ha_commit_trans(this, true))
         result= false;
       options&= ~(OPTION_BEGIN);
-      transaction.all.modified_non_trans_table= false;
       break;
     case COMMIT_RELEASE:
       do_release= 1; /* fall through */
@@ -808,7 +807,6 @@
       if (transaction_services.ha_rollback_trans(this, true))
         result= false;
       options&= ~(OPTION_BEGIN);
-      transaction.all.modified_non_trans_table= false;
       if (result == true && (completion == ROLLBACK_AND_CHAIN))
         result= startTransaction();
       break;
@@ -843,7 +841,6 @@
       result= false;
   }
   options&= ~(OPTION_BEGIN);
-  transaction.all.modified_non_trans_table= false;
   return result;
 }
 
@@ -862,9 +859,7 @@
 
     if (opt == START_TRANS_OPT_WITH_CONS_SNAPSHOT)
     {
-      // TODO make this a loop for all engines, not just this one (Inno only
-      // right now)
-      if (plugin::StorageEngine::startConsistentSnapshot(this))
+      if (plugin::TransactionalStorageEngine::startConsistentSnapshot(this))
       {
         result= false;
       }
@@ -1695,7 +1690,7 @@
 
 int session_non_transactional_update(const Session *session)
 {
-  return(session->transaction.all.modified_non_trans_table);
+  return(session->transaction.all.hasModifiedNonTransData());
 }
 
 void session_mark_transaction_to_rollback(Session *session, bool all)
@@ -1773,15 +1768,6 @@
   server_status&= ~ (SERVER_MORE_RESULTS_EXISTS |
                           SERVER_QUERY_NO_INDEX_USED |
                           SERVER_QUERY_NO_GOOD_INDEX_USED);
-  /*
-    If in autocommit mode and not in a transaction, reset
-    OPTION_STATUS_NO_TRANS_UPDATE to not get warnings
-    in ha_rollback_trans() about some tables couldn't be rolled back.
-  */
-  if (!(options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)))
-  {
-    transaction.all.modified_non_trans_table= false;
-  }
 
   clear_error();
   main_da.reset_diagnostics_area();

=== modified file 'drizzled/session.h'
--- drizzled/session.h	2010-02-18 06:31:05 +0000
+++ drizzled/session.h	2010-02-19 16:11:03 +0000
@@ -25,7 +25,7 @@
 
 #include "drizzled/plugin.h"
 #include <drizzled/sql_locale.h>
-#include <drizzled/ha_trx_info.h>
+#include "drizzled/resource_context.h"
 #include <drizzled/cursor.h>
 #include <drizzled/current_session.h>
 #include <drizzled/sql_error.h>
@@ -34,6 +34,7 @@
 #include <drizzled/xid.h>
 #include "drizzled/query_id.h"
 #include "drizzled/named_savepoint.h"
+#include "drizzled/transaction_context.h"
 
 #include <netdb.h>
 #include <map>
@@ -320,7 +321,7 @@
     this should not be used.
     @sa trans_register_ha()
   */
-  Ha_trx_info ha_info[2];
+  drizzled::ResourceContext resource_context[2];
 
   Ha_data() :ha_ptr(NULL) {}
 };
@@ -539,20 +540,15 @@
   query_id_t warn_query_id;
 public:
   void **getEngineData(const plugin::StorageEngine *engine);
-  Ha_trx_info *getEngineInfo(const plugin::StorageEngine *engine,
-                             size_t index= 0);
+  ResourceContext *getResourceContext(const plugin::StorageEngine *engine,
+                                      size_t index= 0);
 
   struct st_transactions {
-    std::deque<drizzled::NamedSavepoint> savepoints;
-    Session_TRANS all;			// Trans since BEGIN WORK
-    Session_TRANS stmt;			// Trans for current statement
+    std::deque<NamedSavepoint> savepoints;
+    TransactionContext all; ///< Trans since BEGIN WORK
+    TransactionContext stmt; ///< Trans for current statement
     XID_STATE xid_state;
 
-    /*
-       Tables changed in transaction (that must be invalidated in query cache).
-       List contain only transactional tables, that not invalidated in query
-       cache (instead of full list of changed in transaction tables).
-    */
     void cleanup()
     {
       savepoints.clear();
@@ -564,6 +560,7 @@
       xid_state()
     { }
   } transaction;
+
   Field *dup_field;
   sigset_t signals;
 

=== modified file 'drizzled/set_var.cc'
--- drizzled/set_var.cc	2010-02-18 01:10:04 +0000
+++ drizzled/set_var.cc	2010-02-19 16:11:03 +0000
@@ -1379,7 +1379,6 @@
     {
       /* We changed to auto_commit mode */
       session->options&= ~(uint64_t) (OPTION_BEGIN);
-      session->transaction.all.modified_non_trans_table= false;
       session->server_status|= SERVER_STATUS_AUTOCOMMIT;
       TransactionServices &transaction_services= TransactionServices::singleton();
       if (transaction_services.ha_commit_trans(session, true))
@@ -1387,7 +1386,6 @@
     }
     else
     {
-      session->transaction.all.modified_non_trans_table= false;
       session->server_status&= ~SERVER_STATUS_AUTOCOMMIT;
     }
   }

=== modified file 'drizzled/sql_delete.cc'
--- drizzled/sql_delete.cc	2010-02-04 08:14:46 +0000
+++ drizzled/sql_delete.cc	2010-02-06 03:22:59 +0000
@@ -299,15 +299,15 @@
   transactional_table= table->cursor->has_transactions();
 
   if (!transactional_table && deleted > 0)
-    session->transaction.stmt.modified_non_trans_table= true;
+    session->transaction.stmt.markModifiedNonTransData();
 
   /* See similar binlogging code in sql_update.cc, for comments */
-  if ((error < 0) || session->transaction.stmt.modified_non_trans_table)
+  if ((error < 0) || session->transaction.stmt.hasModifiedNonTransData())
   {
-    if (session->transaction.stmt.modified_non_trans_table)
-      session->transaction.all.modified_non_trans_table= true;
+    if (session->transaction.stmt.hasModifiedNonTransData())
+      session->transaction.all.markModifiedNonTransData();
   }
-  assert(transactional_table || !deleted || session->transaction.stmt.modified_non_trans_table);
+  assert(transactional_table || !deleted || session->transaction.stmt.hasModifiedNonTransData());
   free_underlaid_joins(session, select_lex);
 
   DRIZZLE_DELETE_DONE((error >= 0 || session->is_error()), deleted);

=== modified file 'drizzled/sql_insert.cc'
--- drizzled/sql_insert.cc	2010-02-05 03:59:51 +0000
+++ drizzled/sql_insert.cc	2010-02-14 20:26:43 +0000
@@ -29,6 +29,7 @@
 #include "drizzled/sql_table.h"
 #include "drizzled/pthread_globals.h"
 #include "drizzled/transaction_services.h"
+#include "drizzled/plugin/transactional_storage_engine.h"
 
 namespace drizzled
 {
@@ -381,7 +382,7 @@
     }
 
     // Release latches in case bulk insert takes a long time
-    plugin::StorageEngine::releaseTemporaryLatches(session);
+    plugin::TransactionalStorageEngine::releaseTemporaryLatches(session);
 
     error=write_record(session, table ,&info);
     if (error)
@@ -413,12 +414,12 @@
     transactional_table= table->cursor->has_transactions();
 
     changed= (info.copied || info.deleted || info.updated);
-    if ((changed && error <= 0) || session->transaction.stmt.modified_non_trans_table)
+    if ((changed && error <= 0) || session->transaction.stmt.hasModifiedNonTransData())
     {
-      if (session->transaction.stmt.modified_non_trans_table)
-	session->transaction.all.modified_non_trans_table= true;
+      if (session->transaction.stmt.hasModifiedNonTransData())
+	session->transaction.all.markModifiedNonTransData();
     }
-    assert(transactional_table || !changed || session->transaction.stmt.modified_non_trans_table);
+    assert(transactional_table || !changed || session->transaction.stmt.hasModifiedNonTransData());
 
   }
   session->set_proc_info("end");
@@ -696,7 +697,7 @@
     then both on update triggers will work instead. Similarly both on
     delete triggers will be invoked if we will delete conflicting records.
 
-    Sets session->transaction.stmt.modified_non_trans_table to true if table which is updated didn't have
+    Sets session->transaction.stmt.modified_non_trans_data to true if table which is updated didn't have
     transactions.
 
   RETURN VALUE
@@ -895,7 +896,7 @@
             goto err;
           info->deleted++;
           if (!table->cursor->has_transactions())
-            session->transaction.stmt.modified_non_trans_table= true;
+            session->transaction.stmt.markModifiedNonTransData();
           /* Let us attempt do write_row() once more */
         }
       }
@@ -926,7 +927,7 @@
   if (key)
     free(key);
   if (!table->cursor->has_transactions())
-    session->transaction.stmt.modified_non_trans_table= true;
+    session->transaction.stmt.markModifiedNonTransData();
   return(0);
 
 err:
@@ -1253,7 +1254,7 @@
     return(1);
 
   // Release latches in case bulk insert takes a long time
-  plugin::StorageEngine::releaseTemporaryLatches(session);
+  plugin::TransactionalStorageEngine::releaseTemporaryLatches(session);
 
   error= write_record(session, table, &info);
 
@@ -1326,11 +1327,11 @@
       We must invalidate the table in the query cache before binlog writing
       and ha_autocommit_or_rollback.
     */
-    if (session->transaction.stmt.modified_non_trans_table)
-      session->transaction.all.modified_non_trans_table= true;
+    if (session->transaction.stmt.hasModifiedNonTransData())
+      session->transaction.all.markModifiedNonTransData();
   }
   assert(trans_table || !changed ||
-              session->transaction.stmt.modified_non_trans_table);
+              session->transaction.stmt.hasModifiedNonTransData());
 
   table->cursor->ha_release_auto_increment();
 
@@ -1392,7 +1393,7 @@
     changed= (info.copied || info.deleted || info.updated);
     transactional_table= table->cursor->has_transactions();
     assert(transactional_table || !changed ||
-		session->transaction.stmt.modified_non_trans_table);
+		session->transaction.stmt.hasModifiedNonTransData());
     table->cursor->ha_release_auto_increment();
   }
 
@@ -1701,8 +1702,6 @@
 
 void select_create::send_error(uint32_t errcode,const char *err)
 {
-
-
   /*
     This will execute any rollbacks that are necessary before writing
     the transcation cache.
@@ -1754,8 +1753,6 @@
 
 void select_create::abort()
 {
-
-
   /*
     In select_insert::abort() we roll back the statement, including
     truncating the transaction cache of the binary log. To do this, we
@@ -1772,8 +1769,6 @@
     log state.
   */
   select_insert::abort();
-  session->transaction.stmt.modified_non_trans_table= false;
-
 
   if (m_plock)
   {

=== modified file 'drizzled/sql_load.cc'
--- drizzled/sql_load.cc	2010-02-04 08:14:46 +0000
+++ drizzled/sql_load.cc	2010-02-06 03:22:59 +0000
@@ -390,14 +390,14 @@
   sprintf(name, ER(ER_LOAD_INFO), (uint32_t) info.records, (uint32_t) info.deleted,
 	  (uint32_t) (info.records - info.copied), (uint32_t) session->cuted_fields);
 
-  if (session->transaction.stmt.modified_non_trans_table)
-    session->transaction.all.modified_non_trans_table= true;
+  if (session->transaction.stmt.hasModifiedNonTransData())
+    session->transaction.all.markModifiedNonTransData();
 
   /* ok to client sent only after binlog write and engine commit */
   session->my_ok(info.copied + info.deleted, 0, 0L, name);
 err:
   assert(transactional_table || !(info.copied || info.deleted) ||
-              session->transaction.stmt.modified_non_trans_table);
+              session->transaction.stmt.hasModifiedNonTransData());
   table->cursor->ha_release_auto_increment();
   table->auto_increment_field_not_null= false;
   session->abort_on_warning= 0;

=== modified file 'drizzled/sql_parse.cc'
--- drizzled/sql_parse.cc	2010-02-18 06:17:33 +0000
+++ drizzled/sql_parse.cc	2010-02-19 16:11:03 +0000
@@ -506,7 +506,7 @@
 
   status_var_increment(session->status_var.com_stat[lex->sql_command]);
 
-  assert(session->transaction.stmt.modified_non_trans_table == false);
+  assert(session->transaction.stmt.hasModifiedNonTransData() == false);
 
   /* now we are ready to execute the statement */
   res= lex->statement->execute();

=== modified file 'drizzled/sql_update.cc'
--- drizzled/sql_update.cc	2010-02-05 04:05:31 +0000
+++ drizzled/sql_update.cc	2010-02-14 20:26:43 +0000
@@ -511,7 +511,7 @@
   table->cursor->try_semi_consistent_read(0);
 
   if (!transactional_table && updated > 0)
-    session->transaction.stmt.modified_non_trans_table= true;
+    session->transaction.stmt.markModifiedNonTransData();
 
   end_read_record(&info);
   delete select;
@@ -527,12 +527,12 @@
     Sometimes we want to binlog even if we updated no rows, in case user used
     it to be sure master and slave are in same state.
   */
-  if ((error < 0) || session->transaction.stmt.modified_non_trans_table)
+  if ((error < 0) || session->transaction.stmt.hasModifiedNonTransData())
   {
-    if (session->transaction.stmt.modified_non_trans_table)
-      session->transaction.all.modified_non_trans_table= true;
+    if (session->transaction.stmt.hasModifiedNonTransData())
+      session->transaction.all.markModifiedNonTransData();
   }
-  assert(transactional_table || !updated || session->transaction.stmt.modified_non_trans_table);
+  assert(transactional_table || !updated || session->transaction.stmt.hasModifiedNonTransData());
   free_underlaid_joins(session, select_lex);
 
   /* If LAST_INSERT_ID(X) was used, report X */

=== modified file 'drizzled/statement/rollback_to_savepoint.cc'
--- drizzled/statement/rollback_to_savepoint.cc	2010-01-28 20:48:31 +0000
+++ drizzled/statement/rollback_to_savepoint.cc	2010-02-06 03:22:59 +0000
@@ -75,7 +75,7 @@
       /* Found the named savepoint we want to rollback to */
       (void) transaction_services.ha_rollback_to_savepoint(session, first_savepoint);
 
-      if (session->transaction.all.modified_non_trans_table)
+      if (session->transaction.all.hasModifiedNonTransData())
       {
         push_warning(session, 
                      DRIZZLE_ERROR::WARN_LEVEL_WARN,
@@ -126,7 +126,7 @@
   }
   if (found)
   {
-    if (session->transaction.all.modified_non_trans_table)
+    if (session->transaction.all.hasModifiedNonTransData())
     {
       push_warning(session, 
                    DRIZZLE_ERROR::WARN_LEVEL_WARN,

=== modified file 'drizzled/table.cc'
--- drizzled/table.cc	2010-02-10 23:07:02 +0000
+++ drizzled/table.cc	2010-02-14 20:26:43 +0000
@@ -29,6 +29,7 @@
 #include <drizzled/gettext.h>
 
 #include "drizzled/plugin/info_schema_table.h"
+#include "drizzled/plugin/transactional_storage_engine.h"
 #include <drizzled/nested_join.h>
 #include <drizzled/sql_parse.h>
 #include <drizzled/item/sum.h>
@@ -3313,7 +3314,7 @@
   session->set_proc_info("removing tmp table");
 
   // Release latches since this can take a long time
-  plugin::StorageEngine::releaseTemporaryLatches(session);
+  plugin::TransactionalStorageEngine::releaseTemporaryLatches(session);
 
   if (cursor)
   {
@@ -3357,7 +3358,7 @@
   }
 
   // Release latches since this can take a long time
-  plugin::StorageEngine::releaseTemporaryLatches(session);
+  plugin::TransactionalStorageEngine::releaseTemporaryLatches(session);
 
   new_table= *table;
   share= *table->s;

=== added file 'drizzled/transaction_context.h'
--- drizzled/transaction_context.h	1970-01-01 00:00:00 +0000
+++ drizzled/transaction_context.h	2010-02-06 03:22:59 +0000
@@ -0,0 +1,86 @@
+/* -*- mode: c++; c-basic-offset: 2; indent-tabs-mode: nil; -*-
+ *  vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
+ *
+ *  Copyright (C) 2008 Sun Microsystems
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; version 2 of the License.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ */
+
+#ifndef DRIZZLED_TRANSACTION_CONTEXT_H
+#define DRIZZLED_TRANSACTION_CONTEXT_H
+
+#include <vector>
+
+namespace drizzled
+{
+class ResourceContext;
+
+class TransactionContext
+{
+public:
+  TransactionContext() :
+    no_2pc(false),
+    resource_contexts(),
+    modified_non_trans_data(false)
+  {}
+
+  void reset() { no_2pc= false; modified_non_trans_data= false; resource_contexts.clear();}
+
+  typedef std::vector<ResourceContext *> ResourceContexts;
+
+  void setResourceContexts(ResourceContexts &new_contexts)
+  {
+    resource_contexts.assign(new_contexts.begin(), new_contexts.end());
+  }
+
+  ResourceContexts &getResourceContexts()
+  {
+    return resource_contexts;
+  }
+  /** Register a resource context in this transaction context */
+  void registerResource(ResourceContext *resource)
+  {
+    resource_contexts.push_back(resource);
+  }
+
+  /**
+   * Marks that this transaction has modified state
+   * of some non-transactional data.
+   */
+  void markModifiedNonTransData()
+  {
+    modified_non_trans_data= true;
+  }
+
+  /**
+   * Returns true if the transaction has modified
+   * state of some non-transactional data.
+   */
+  bool hasModifiedNonTransData() const
+  {
+    return modified_non_trans_data;
+  }
+
+  /* true is not all entries in the resource contexts support 2pc */
+  bool no_2pc;
+private:
+  /** Resource that registered in this transaction */
+  ResourceContexts resource_contexts;
+  /** Whether this transaction has changed non-transaction data state */
+  bool modified_non_trans_data;
+};
+
+} /* namespace drizzled */
+
+#endif /* DRIZZLED_TRANSACTION_CONTEXT_H */

=== modified file 'drizzled/transaction_services.cc'
--- drizzled/transaction_services.cc	2010-02-04 08:14:46 +0000
+++ drizzled/transaction_services.cc	2010-02-14 19:27:57 +0000
@@ -31,15 +31,22 @@
 #include "drizzled/sql_base.h"
 #include "drizzled/replication_services.h"
 #include "drizzled/transaction_services.h"
+#include "drizzled/transaction_context.h"
+#include "drizzled/resource_context.h"
 #include "drizzled/lock.h"
 #include "drizzled/item/int.h"
 #include "drizzled/item/empty_string.h"
 #include "drizzled/field/timestamp.h"
 #include "drizzled/plugin/client.h"
+#include "drizzled/plugin/xa_storage_engine.h"
 #include "drizzled/internal/my_sys.h"
 
 using namespace std;
 
+#include <vector>
+#include <algorithm>
+#include <functional>
+
 namespace drizzled
 {
 
@@ -143,7 +150,7 @@
 
   The server stores its transaction-related data in
   session->transaction. This structure has two members of type
-  Session_TRANS. These members correspond to the statement and
+  TransactionContext. These members correspond to the statement and
   normal transactions respectively:
 
   - session->transaction.stmt contains a list of engines
@@ -346,10 +353,10 @@
     times per transaction.
 
 */
-void TransactionServices::trans_register_ha(Session *session, bool all, plugin::StorageEngine *engine)
+void TransactionServices::trans_register_ha(Session *session, bool all, plugin::TransactionalStorageEngine *engine)
 {
-  Session_TRANS *trans;
-  Ha_trx_info *ha_info;
+  TransactionContext *trans;
+  ResourceContext *resource_context;
 
   if (all)
   {
@@ -359,14 +366,15 @@
   else
     trans= &session->transaction.stmt;
 
-  ha_info= session->getEngineInfo(engine, all ? 1 : 0);
+  resource_context= session->getResourceContext(engine, all ? 1 : 0);
 
-  if (ha_info->is_started())
+  if (resource_context->isStarted())
     return; /* already registered, return */
 
-  ha_info->register_ha(trans, engine);
+  resource_context->setResource(engine);
+  trans->registerResource(resource_context);
 
-  trans->no_2pc|= not engine->has_2pc();
+  trans->no_2pc|= not engine->hasTwoPhaseCommit();
   if (session->transaction.xid_state.xid.is_null())
     session->transaction.xid_state.xid.set(session->getQueryId());
 }
@@ -386,33 +394,37 @@
 */
 static
 bool
-ha_check_and_coalesce_trx_read_only(Session *session, Ha_trx_info *ha_list,
-                                    bool all)
+ha_check_and_coalesce_trx_read_only(Session *session,
+                                    TransactionContext::ResourceContexts &resource_contexts,
+                                    bool normal_transaction)
 {
   /* The number of storage engines that have actual changes. */
-  unsigned rw_ha_count= 0;
-  Ha_trx_info *ha_info;
+  unsigned num_resources_modified_data= 0;
+  ResourceContext *resource_context;
 
-  for (ha_info= ha_list; ha_info; ha_info= ha_info->next())
+  for (TransactionContext::ResourceContexts::iterator it= resource_contexts.begin();
+       it != resource_contexts.end();
+       ++it)
   {
-    if (ha_info->is_trx_read_write())
-      ++rw_ha_count;
+    resource_context= *it;
+    if (resource_context->hasModifiedData())
+      ++num_resources_modified_data;
 
-    if (! all)
+    if (! normal_transaction)
     {
-      Ha_trx_info *ha_info_all= session->getEngineInfo(ha_info->engine(), 1);
-      assert(ha_info != ha_info_all);
+      ResourceContext *resource_context_normal= session->getResourceContext(resource_context->getResource(), true);
+      assert(resource_context != resource_context_normal);
       /*
         Merge read-only/read-write information about statement
         transaction to its enclosing normal transaction. Do this
         only if in a real transaction -- that is, if we know
-        that ha_info_all is registered in session->transaction.all.
+        that resource_context_all is registered in session->transaction.all.
         Since otherwise we only clutter the normal transaction flags.
       */
-      if (ha_info_all->is_started()) /* false if autocommit. */
-        ha_info_all->coalesce_trx_with(ha_info);
+      if (resource_context_normal->isStarted()) /* false if autocommit. */
+        resource_context_normal->coalesceWith(resource_context);
     }
-    else if (rw_ha_count > 1)
+    else if (num_resources_modified_data > 1)
     {
       /*
         It is a normal transaction, so we don't need to merge read/write
@@ -422,7 +434,7 @@
       break;
     }
   }
-  return rw_ha_count > 1;
+  return num_resources_modified_data > 1;
 }
 
 
@@ -440,16 +452,17 @@
     stored functions or triggers. So we simply do nothing now.
     TODO: This should be fixed in later ( >= 5.1) releases.
 */
-int TransactionServices::ha_commit_trans(Session *session, bool all)
+int TransactionServices::ha_commit_trans(Session *session, bool normal_transaction)
 {
   int error= 0, cookie= 0;
   /*
     'all' means that this is either an explicit commit issued by
     user, or an implicit commit issued by a DDL.
   */
-  Session_TRANS *trans= all ? &session->transaction.all : &session->transaction.stmt;
-  bool is_real_trans= all || session->transaction.all.ha_list == 0;
-  Ha_trx_info *ha_info= trans->ha_list;
+  TransactionContext *trans= normal_transaction ? &session->transaction.all : &session->transaction.stmt;
+  TransactionContext::ResourceContexts &resource_contexts= trans->getResourceContexts();
+
+  bool is_real_trans= normal_transaction || session->transaction.all.getResourceContexts().empty();
 
   /*
     We must not commit the normal transaction if a statement
@@ -457,39 +470,39 @@
     flags will not get propagated to its normal transaction's
     counterpart.
   */
-  assert(session->transaction.stmt.ha_list == NULL ||
+  assert(session->transaction.stmt.getResourceContexts().empty() ||
               trans == &session->transaction.stmt);
 
-  if (ha_info)
+  if (resource_contexts.empty() == false)
   {
     bool must_2pc;
 
     if (is_real_trans && wait_if_global_read_lock(session, 0, 0))
     {
-      ha_rollback_trans(session, all);
+      ha_rollback_trans(session, normal_transaction);
       return 1;
     }
 
-    must_2pc= ha_check_and_coalesce_trx_read_only(session, ha_info, all);
+    must_2pc= ha_check_and_coalesce_trx_read_only(session, resource_contexts, normal_transaction);
 
-    if (!trans->no_2pc && must_2pc)
+    if (! trans->no_2pc && must_2pc)
     {
-      for (; ha_info && !error; ha_info= ha_info->next())
+      for (TransactionContext::ResourceContexts::iterator it= resource_contexts.begin();
+           it != resource_contexts.end() && ! error;
+           ++it)
       {
+        ResourceContext *resource_context= *it;
         int err;
-        plugin::StorageEngine *engine= ha_info->engine();
         /*
           Do not call two-phase commit if this particular
           transaction is read-only. This allows for simpler
           implementation in engines that are always read-only.
         */
-        if (! ha_info->is_trx_read_write())
+        if (! resource_context->hasModifiedData())
           continue;
-        /*
-          Sic: we know that prepare() is not NULL since otherwise
-          trans->no_2pc would have been set.
-        */
-        if ((err= engine->prepare(session, all)))
+
+        plugin::StorageEngine *engine= resource_context->getResource();
+        if ((err= static_cast<plugin::XaStorageEngine *>(engine)->prepare(session, normal_transaction)))
         {
           my_error(ER_ERROR_DURING_COMMIT, MYF(0), err);
           error= 1;
@@ -498,12 +511,12 @@
       }
       if (error)
       {
-        ha_rollback_trans(session, all);
+        ha_rollback_trans(session, normal_transaction);
         error= 1;
         goto end;
       }
     }
-    error=ha_commit_one_phase(session, all) ? (cookie ? 2 : 1) : 0;
+    error= ha_commit_one_phase(session, normal_transaction) ? (cookie ? 2 : 1) : 0;
 end:
     if (is_real_trans)
       start_waiting_global_read_lock(session);
@@ -515,34 +528,39 @@
   @note
   This function does not care about global read lock. A caller should.
 */
-int TransactionServices::ha_commit_one_phase(Session *session, bool all)
+int TransactionServices::ha_commit_one_phase(Session *session, bool normal_transaction)
 {
   int error=0;
-  Session_TRANS *trans=all ? &session->transaction.all : &session->transaction.stmt;
-  bool is_real_trans=all || session->transaction.all.ha_list == 0;
-  Ha_trx_info *ha_info= trans->ha_list, *ha_info_next;
-  if (ha_info)
+  TransactionContext *trans= normal_transaction ? &session->transaction.all : &session->transaction.stmt;
+  TransactionContext::ResourceContexts &resource_contexts= trans->getResourceContexts();
+
+  bool is_real_trans= normal_transaction || session->transaction.all.getResourceContexts().empty();
+
+  if (resource_contexts.empty() == false)
   {
-    for (; ha_info; ha_info= ha_info_next)
+    for (TransactionContext::ResourceContexts::iterator it= resource_contexts.begin();
+         it != resource_contexts.end();
+         ++it)
     {
       int err;
-      plugin::StorageEngine *engine= ha_info->engine();
-      if ((err= engine->commit(session, all)))
+      ResourceContext *resource_context= *it;
+      plugin::TransactionalStorageEngine *engine= static_cast<plugin::TransactionalStorageEngine *>(resource_context->getResource());
+      if ((err= engine->commit(session, normal_transaction)))
       {
         my_error(ER_ERROR_DURING_COMMIT, MYF(0), err);
-        error=1;
+        error= 1;
       }
       status_var_increment(session->status_var.ha_commit_count);
-      ha_info_next= ha_info->next();
-      ha_info->reset(); /* keep it conveniently zero-filled */
+      resource_context->reset(); /* keep it conveniently zero-filled */
     }
-    trans->ha_list= 0;
-    trans->no_2pc=0;
+    trans->reset();
+
     if (is_real_trans)
       session->transaction.xid_state.xid.null();
-    if (all)
+
+    if (normal_transaction)
     {
-      session->variables.tx_isolation=session->session_tx_isolation;
+      session->variables.tx_isolation= session->session_tx_isolation;
       session->transaction.cleanup();
     }
   }
@@ -551,9 +569,9 @@
     if (is_real_trans)
     {
       /* 
-        * We commit the normal transaction by finalizing the transaction message
-        * and propogating the message to all registered replicators.
-        */
+       * We commit the normal transaction by finalizing the transaction message
+       * and propogating the message to all registered replicators.
+       */
       ReplicationServices &replication_services= ReplicationServices::singleton();
       replication_services.commitTransaction(session);
     }
@@ -561,38 +579,39 @@
   return error;
 }
 
-
-int TransactionServices::ha_rollback_trans(Session *session, bool all)
+int TransactionServices::ha_rollback_trans(Session *session, bool normal_transaction)
 {
-  int error=0;
-  Session_TRANS *trans=all ? &session->transaction.all : &session->transaction.stmt;
-  Ha_trx_info *ha_info= trans->ha_list, *ha_info_next;
-  bool is_real_trans=all || session->transaction.all.ha_list == 0;
+  int error= 0;
+  TransactionContext *trans= normal_transaction ? &session->transaction.all : &session->transaction.stmt;
+  TransactionContext::ResourceContexts &resource_contexts= trans->getResourceContexts();
+
+  bool is_real_trans= normal_transaction || session->transaction.all.getResourceContexts().empty();
 
   /*
     We must not rollback the normal transaction if a statement
     transaction is pending.
   */
-  assert(session->transaction.stmt.ha_list == NULL ||
+  assert(session->transaction.stmt.getResourceContexts().empty() ||
               trans == &session->transaction.stmt);
 
-  if (ha_info)
+  if (resource_contexts.empty() == false)
   {
-    for (; ha_info; ha_info= ha_info_next)
+    for (TransactionContext::ResourceContexts::iterator it= resource_contexts.begin();
+         it != resource_contexts.end();
+         ++it)
     {
       int err;
-      plugin::StorageEngine *engine= ha_info->engine();
-      if ((err= engine->rollback(session, all)))
+      ResourceContext *resource_context= *it;
+      plugin::TransactionalStorageEngine *engine= static_cast<plugin::TransactionalStorageEngine *>(resource_context->getResource());
+      if ((err= engine->rollback(session, normal_transaction)))
       { // cannot happen
         my_error(ER_ERROR_DURING_ROLLBACK, MYF(0), err);
         error=1;
       }
       status_var_increment(session->status_var.ha_rollback_count);
-      ha_info_next= ha_info->next();
-      ha_info->reset(); /* keep it conveniently zero-filled */
+      resource_context->reset(); /* keep it conveniently zero-filled */
     }
-    trans->ha_list= 0;
-    trans->no_2pc=0;
+    trans->reset();
     
     /* 
      * We need to signal the ROLLBACK to ReplicationServices here
@@ -606,13 +625,13 @@
 
     if (is_real_trans)
       session->transaction.xid_state.xid.null();
-    if (all)
+    if (normal_transaction)
     {
       session->variables.tx_isolation=session->session_tx_isolation;
       session->transaction.cleanup();
     }
   }
-  if (all)
+  if (normal_transaction)
     session->transaction_rollback_request= false;
 
   /*
@@ -624,10 +643,14 @@
     the error log; but we don't want users to wonder why they have this
     message in the error log, so we don't send it.
   */
-  if (is_real_trans && session->transaction.all.modified_non_trans_table && session->killed != Session::KILL_CONNECTION)
+  if (is_real_trans &&
+      session->transaction.all.hasModifiedNonTransData() &&
+      session->killed != Session::KILL_CONNECTION)
+  {
     push_warning(session, DRIZZLE_ERROR::WARN_LEVEL_WARN,
                  ER_WARNING_NOT_COMPLETE_ROLLBACK,
                  ER(ER_WARNING_NOT_COMPLETE_ROLLBACK));
+  }
   return error;
 }
 
@@ -644,9 +667,9 @@
 */
 int TransactionServices::ha_autocommit_or_rollback(Session *session, int error)
 {
-  if (session->transaction.stmt.ha_list)
+  if (session->transaction.stmt.getResourceContexts().empty() == false)
   {
-    if (!error)
+    if (! error)
     {
       if (ha_commit_trans(session, false))
         error= 1;
@@ -658,9 +681,8 @@
         (void) ha_rollback_trans(session, true);
     }
 
-    session->variables.tx_isolation=session->session_tx_isolation;
+    session->variables.tx_isolation= session->session_tx_isolation;
   }
-
   return error;
 }
 
@@ -709,50 +731,86 @@
   return 0;
 }
 
+struct ResourceContextCompare : public std::binary_function<ResourceContext *, ResourceContext *, bool>
+{
+  result_type operator()(const ResourceContext *lhs, const ResourceContext *rhs) const
+  {
+    return lhs->getResource()->getSlot() < rhs->getResource()->getSlot();
+  }
+};
 
 int TransactionServices::ha_rollback_to_savepoint(Session *session, NamedSavepoint &sv)
 {
   int error= 0;
-  Session_TRANS *trans= &session->transaction.all;
-  Ha_trx_info *ha_info, *ha_info_next;
+  TransactionContext *trans= &session->transaction.all;
+  TransactionContext::ResourceContexts &tran_resource_contexts= trans->getResourceContexts();
+  TransactionContext::ResourceContexts &sv_resource_contexts= sv.getResourceContexts();
 
-  trans->no_2pc=0;
+  trans->no_2pc= false;
   /*
     rolling back to savepoint in all storage engines that were part of the
     transaction when the savepoint was set
   */
-  for (ha_info= sv.ha_list; ha_info; ha_info= ha_info->next())
+  for (TransactionContext::ResourceContexts::iterator it= sv_resource_contexts.begin();
+       it != sv_resource_contexts.end();
+       ++it)
   {
     int err;
-    plugin::StorageEngine *engine= ha_info->engine();
-    assert(engine);
-    if ((err= engine->savepoint_rollback(session, sv)))
+    ResourceContext *resource_context= *it;
+    plugin::TransactionalStorageEngine *engine= static_cast<plugin::TransactionalStorageEngine *>(resource_context->getResource());
+    assert(engine != NULL);
+    if ((err= engine->rollbackToSavepoint(session, sv)))
     { // cannot happen
       my_error(ER_ERROR_DURING_ROLLBACK, MYF(0), err);
       error= 1;
     }
     status_var_increment(session->status_var.ha_savepoint_rollback_count);
-    trans->no_2pc|= not engine->has_2pc();
+    trans->no_2pc|= not engine->hasTwoPhaseCommit();
   }
   /*
     rolling back the transaction in all storage engines that were not part of
     the transaction when the savepoint was set
   */
-  for (ha_info= trans->ha_list; ha_info != sv.ha_list;
-       ha_info= ha_info_next)
   {
-    int err;
-    plugin::StorageEngine *engine= ha_info->engine();
-    if ((err= engine->rollback(session, !(0))))
-    { // cannot happen
-      my_error(ER_ERROR_DURING_ROLLBACK, MYF(0), err);
-      error= 1;
+    TransactionContext::ResourceContexts sorted_tran_resource_contexts(tran_resource_contexts);
+    TransactionContext::ResourceContexts sorted_sv_resource_contexts(sv_resource_contexts);
+    TransactionContext::ResourceContexts set_difference_contexts;
+
+    sort(sorted_tran_resource_contexts.begin(),
+         sorted_tran_resource_contexts.end(),
+         ResourceContextCompare());
+    sort(sorted_sv_resource_contexts.begin(),
+         sorted_sv_resource_contexts.end(),
+         ResourceContextCompare());
+    set_difference(sorted_tran_resource_contexts.begin(),
+                   sorted_tran_resource_contexts.end(),
+                   sorted_sv_resource_contexts.begin(),
+                   sorted_sv_resource_contexts.end(),
+                   set_difference_contexts.begin(),
+                   ResourceContextCompare());
+    /* 
+     * set_difference_contexts now contains all resource contexts
+     * which are in the transaction context but were NOT in the
+     * savepoint's resource contexts.
+     */
+        
+    for (TransactionContext::ResourceContexts::iterator it= set_difference_contexts.begin();
+         it != set_difference_contexts.end();
+         ++it)
+    {
+      ResourceContext *resource_context= *it;
+      int err;
+      plugin::TransactionalStorageEngine *engine= static_cast<plugin::TransactionalStorageEngine *>(resource_context->getResource());
+      if ((err= engine->rollback(session, !(0))))
+      { // cannot happen
+        my_error(ER_ERROR_DURING_ROLLBACK, MYF(0), err);
+        error= 1;
+      }
+      status_var_increment(session->status_var.ha_rollback_count);
+      resource_context->reset(); /* keep it conveniently zero-filled */
     }
-    status_var_increment(session->status_var.ha_rollback_count);
-    ha_info_next= ha_info->next();
-    ha_info->reset(); /* keep it conveniently zero-filled */
   }
-  trans->ha_list= sv.ha_list;
+  trans->setResourceContexts(sv_resource_contexts);
   return error;
 }
 
@@ -765,48 +823,49 @@
 int TransactionServices::ha_savepoint(Session *session, NamedSavepoint &sv)
 {
   int error= 0;
-  Session_TRANS *trans= &session->transaction.all;
-  Ha_trx_info *ha_info= trans->ha_list;
-  for (; ha_info; ha_info= ha_info->next())
+  TransactionContext *trans= &session->transaction.all;
+  TransactionContext::ResourceContexts &resource_contexts= trans->getResourceContexts();
+
+  if (resource_contexts.empty() == false)
   {
-    int err;
-    plugin::StorageEngine *engine= ha_info->engine();
-    assert(engine);
-#ifdef NOT_IMPLEMENTED /*- TODO (examine this againt the original code base) */
-    if (! engine->savepoint_set)
+    for (TransactionContext::ResourceContexts::iterator it= resource_contexts.begin();
+         it != resource_contexts.end();
+         ++it)
     {
-      my_error(ER_CHECK_NOT_IMPLEMENTED, MYF(0), "NamedSavepoint");
-      error= 1;
-      break;
-    } 
-#endif
-    if ((err= engine->savepoint_set(session, sv)))
-    { // cannot happen
-      my_error(ER_GET_ERRNO, MYF(0), err);
-      error= 1;
+      ResourceContext *resource_context= *it;
+      int err;
+      plugin::TransactionalStorageEngine *engine= static_cast<plugin::TransactionalStorageEngine *>(resource_context->getResource());
+      assert(engine);
+      if ((err= engine->setSavepoint(session, sv)))
+      { // cannot happen
+        my_error(ER_GET_ERRNO, MYF(0), err);
+        error= 1;
+      }
+      status_var_increment(session->status_var.ha_savepoint_count);
     }
-    status_var_increment(session->status_var.ha_savepoint_count);
   }
   /*
-    Remember the list of registered storage engines. All new
-    engines are prepended to the beginning of the list.
+    Remember the list of registered storage engines.
   */
-  sv.ha_list= trans->ha_list;
+  sv.setResourceContexts(resource_contexts);
   return error;
 }
 
 int TransactionServices::ha_release_savepoint(Session *session, NamedSavepoint &sv)
 {
   int error= 0;
-  Ha_trx_info *ha_info= sv.ha_list;
-
-  for (; ha_info; ha_info= ha_info->next())
+
+  TransactionContext::ResourceContexts &resource_contexts= sv.getResourceContexts();
+
+  for (TransactionContext::ResourceContexts::iterator it= resource_contexts.begin();
+       it != resource_contexts.end();
+       ++it)
   {
     int err;
-    plugin::StorageEngine *engine= ha_info->engine();
-    /* Savepoint life time is enclosed into transaction life time. */
+    ResourceContext *resource_context= *it;
+    plugin::TransactionalStorageEngine *engine= static_cast<plugin::TransactionalStorageEngine *>(resource_context->getResource());
     assert(engine);
-    if ((err= engine->savepoint_release(session, sv)))
+    if ((err= engine->releaseSavepoint(session, sv)))
     { // cannot happen
       my_error(ER_GET_ERRNO, MYF(0), err);
       error= 1;

=== modified file 'drizzled/transaction_services.h'
--- drizzled/transaction_services.h	2010-02-04 08:14:46 +0000
+++ drizzled/transaction_services.h	2010-02-14 19:27:57 +0000
@@ -30,7 +30,7 @@
 /* some forward declarations needed */
 namespace plugin
 {
-  class StorageEngine;
+  class TransactionalStorageEngine;
 }
 
 class Session;
@@ -72,7 +72,7 @@
   bool mysql_xa_recover(Session *session);
 
   /* these are called by storage engines */
-  void trans_register_ha(Session *session, bool all, plugin::StorageEngine *engine);
+  void trans_register_ha(Session *session, bool all, plugin::TransactionalStorageEngine *engine);
 };
 
 } /* namespace drizzled */

=== modified file 'plugin/innobase/handler/ha_innodb.cc'
--- plugin/innobase/handler/ha_innodb.cc	2010-02-18 22:38:40 +0000
+++ plugin/innobase/handler/ha_innodb.cc	2010-02-19 16:11:03 +0000
@@ -80,7 +80,7 @@
 #include "drizzled/field/blob.h"
 #include "drizzled/field/varstring.h"
 #include "drizzled/field/timestamp.h"
-#include "drizzled/plugin/storage_engine.h"
+#include "drizzled/plugin/xa_storage_engine.h"
 #include "drizzled/plugin/info_schema_table.h"
 #include "drizzled/memory/multi_malloc.h"
 #include "drizzled/pthread_globals.h"
@@ -165,7 +165,7 @@
 # define EQ_CURRENT_SESSION(session) ((session) == current_session)
 #endif /* MYSQL_DYNAMIC_PLUGIN && __WIN__ */
 
-static plugin::StorageEngine* innodb_engine_ptr= NULL;
+static plugin::XaStorageEngine* innodb_engine_ptr= NULL;
 static plugin::TableFunction* status_table_function_ptr= NULL;
 
 static const long AUTOINC_OLD_STYLE_LOCKING = 0;
@@ -253,11 +253,11 @@
 static INNOBASE_SHARE *get_share(const char *table_name);
 static void free_share(INNOBASE_SHARE *share);
 
-class InnobaseEngine : public plugin::StorageEngine
+class InnobaseEngine : public plugin::XaStorageEngine
 {
 public:
   InnobaseEngine(string name_arg) :
-    plugin::StorageEngine(name_arg,
+    plugin::XaStorageEngine(name_arg,
                           HTON_NULL_IN_KEY |
                           HTON_CAN_INDEX_BLOBS |
                           HTON_PRIMARY_KEY_REQUIRED_FOR_POSITION |
@@ -278,20 +278,20 @@
 	Session*	session);	/* in: handle to the MySQL thread of the user
 			whose resources should be free'd */
 
-  virtual int savepoint_set_hook(Session* session,
+  virtual int doSetSavepoint(Session* session,
                                  drizzled::NamedSavepoint &savepoint);
-  virtual int savepoint_rollback_hook(Session* session,
-                                     drizzled::NamedSavepoint &savepoint);
-  virtual int savepoint_release_hook(Session* session,
-                                     drizzled::NamedSavepoint &savepoint);
-  virtual int commit(Session* session, bool all);
-  virtual int rollback(Session* session, bool all);
+  virtual int doRollbackToSavepoint(Session* session,
+                                     drizzled::NamedSavepoint &savepoint);
+  virtual int doReleaseSavepoint(Session* session,
+                                     drizzled::NamedSavepoint &savepoint);
+  virtual int doCommit(Session* session, bool all);
+  virtual int doRollback(Session* session, bool all);
 
   /***********************************************************************
   This function is used to prepare X/Open XA distributed transaction   */
   virtual
   int
-  prepare(
+  doPrepare(
   /*================*/
   			/* out: 0 or error number */
   	Session*	session,	/* in: handle to the MySQL thread of the user
@@ -302,18 +302,18 @@
   This function is used to recover X/Open XA distributed transactions   */
   virtual
   int
-  recover(
+  doRecover(
   /*================*/
   				/* out: number of prepared transactions
   				stored in xid_list */
   	::drizzled::XID*	xid_list,	/* in/out: prepared transactions */
-  	uint	len);		/* in: number of slots in xid_list */
+  	size_t len);		/* in: number of slots in xid_list */
   /***********************************************************************
   This function is used to commit one X/Open XA distributed transaction
   which is in the prepared state */
   virtual
   int
-  commit_by_xid(
+  doCommitXid(
   /*===================*/
   			/* out: 0 or error number */
   	::drizzled::XID*	xid);	/* in: X/Open XA transaction identification */
@@ -322,7 +322,7 @@
   which is in the prepared state */
   virtual
   int
-  rollback_by_xid(
+  doRollbackXid(
   /*=====================*/
   			/* out: 0 or error number */
   	::drizzled::XID	*xid);	/* in: X/Open XA transaction identification */
@@ -352,7 +352,7 @@
   have one. */
   virtual
   int
-  start_consistent_snapshot(
+  doStartConsistentSnapshot(
   /*====================================*/
   			/* out: 0 */
   	Session*	session);	/* in: MySQL thread handle of the user for whom
@@ -379,7 +379,7 @@
 
   virtual
   int
-  release_temporary_latches(
+  doReleaseTemporaryLatches(
   /*===============================*/
 				/* out: 0 */
 	Session*		session);	/* in: MySQL thread */
@@ -818,7 +818,7 @@
 documentation, see Cursor.cc.
 @return	0 */
 int
-InnobaseEngine::release_temporary_latches(
+InnobaseEngine::doReleaseTemporaryLatches(
 /*===============================*/
 	Session*		session)	/*!< in: MySQL thread */
 {
@@ -1512,7 +1512,7 @@
 void
 innobase_register_stmt(
 /*===================*/
-        plugin::StorageEngine*	engine,	/*!< in: Innobase hton */
+        plugin::TransactionalStorageEngine*	engine,	/*!< in: Innobase hton */
 	Session*	session)	/*!< in: MySQL thd (connection) object */
 {
 	assert(engine == innodb_engine_ptr);
@@ -1532,7 +1532,7 @@
 void
 innobase_register_trx_and_stmt(
 /*===========================*/
-        plugin::StorageEngine *engine, /*!< in: Innobase StorageEngine */
+        plugin::TransactionalStorageEngine *engine, /*!< in: Innobase StorageEngine */
 	Session*	session)	/*!< in: MySQL thd (connection) object */
 {
 	/* NOTE that actually innobase_register_stmt() registers also
@@ -1750,7 +1750,7 @@
 
 	if (prebuilt->trx->active_trans == 0) {
 
-		innobase_register_trx_and_stmt(engine, user_session);
+		innobase_register_trx_and_stmt(innodb_engine_ptr, user_session);
 
 		prebuilt->trx->active_trans = 1;
 	}
@@ -2172,7 +2172,7 @@
 have one.
 @return	0 */
 int
-InnobaseEngine::start_consistent_snapshot(
+InnobaseEngine::doStartConsistentSnapshot(
 /*====================================*/
 	Session*	session)	/*!< in: MySQL thread handle of the user for whom
 			the transaction should be committed */
@@ -2214,7 +2214,7 @@
 ended.
 @return	0 */
 int
-InnobaseEngine::commit(
+InnobaseEngine::doCommit(
 /*============*/
 	Session* 	session,	/*!< in: MySQL thread handle of the user for whom
 			the transaction should be committed */
@@ -2239,7 +2239,7 @@
 	1. ::external_lock(),
 	2. ::start_stmt(),
 	3. innobase_query_caching_of_table_permitted(),
-	4. InnobaseEngine::savepoint_set(),
+	4. InnobaseEngine::setSavepoint(),
 	5. ::init_table_handle_for_HANDLER(),
 	6. InnobaseEngine::start_consistent_snapshot(),
 
@@ -2344,7 +2344,7 @@
 Rolls back a transaction or the latest SQL statement.
 @return	0 or error number */
 int
-InnobaseEngine::rollback(
+InnobaseEngine::doRollback(
 /*==============*/
 	Session*	session,/*!< in: handle to the MySQL thread of the user
 			whose transaction should be rolled back */
@@ -2415,7 +2415,7 @@
 @return 0 if success, HA_ERR_NO_SAVEPOINT if no savepoint with the
 given name */
 int
-InnobaseEngine::savepoint_rollback_hook(
+InnobaseEngine::doRollbackToSavepoint(
 /*===========================*/
 	Session*	session,		/*!< in: handle to the MySQL thread of the user
 				whose transaction should be rolled back */
@@ -2446,7 +2446,7 @@
 @return 0 if success, HA_ERR_NO_SAVEPOINT if no savepoint with the
 given name */
 int
-InnobaseEngine::savepoint_release_hook(
+InnobaseEngine::doReleaseSavepoint(
 /*=======================*/
 	Session*	session,		/*!< in: handle to the MySQL thread of the user
 				whose transaction should be rolled back */
@@ -2469,7 +2469,7 @@
 Sets a transaction savepoint.
 @return	always 0, that is, always succeeds */
 int
-InnobaseEngine::savepoint_set_hook(
+InnobaseEngine::doSetSavepoint(
 /*===============*/
 	Session*	session,/*!< in: handle to the MySQL thread */
 	drizzled::NamedSavepoint &named_savepoint)	/*!< in: savepoint data */
@@ -2774,7 +2774,7 @@
 	holding btr_search_latch. This breaks the latching order as
 	we acquire dict_sys->mutex below and leads to a deadlock. */
 	if (session != NULL) {
-		engine->release_temporary_latches(session);
+		getTransactionalEngine()->releaseTemporaryLatches(session);
 	}
 
 	normalize_table_name(norm_name, name);
@@ -2989,7 +2989,7 @@
 
 	session = ha_session();
 	if (session != NULL) {
-		engine->release_temporary_latches(session);
+		getTransactionalEngine()->releaseTemporaryLatches(session);
 	}
 
 	row_prebuilt_free(prebuilt, FALSE);
@@ -3940,7 +3940,7 @@
 			no need to re-acquire locks on it. */
 
 			/* Altering to InnoDB format */
-			engine->commit(user_session, 1);
+			getTransactionalEngine()->commit(user_session, 1);
 			/* Note that this transaction is still active. */
 			prebuilt->trx->active_trans = 1;
 			/* We will need an IX lock on the destination table. */
@@ -3956,7 +3956,7 @@
 
 			/* Commit the transaction.  This will release the table
 			locks, so they have to be acquired again. */
-			engine->commit(user_session, 1);
+			getTransactionalEngine()->commit(user_session, 1);
 			/* Note that this transaction is still active. */
 			prebuilt->trx->active_trans = 1;
 			/* Re-acquire the table lock on the source table. */
@@ -7252,10 +7252,10 @@
 	/* Set the MySQL flag to mark that there is an active transaction */
 	if (trx->active_trans == 0) {
 
-		innobase_register_trx_and_stmt(engine, session);
+		innobase_register_trx_and_stmt(innodb_engine_ptr, session);
 		trx->active_trans = 1;
 	} else {
-		innobase_register_stmt(engine, session);
+		innobase_register_stmt(innodb_engine_ptr, session);
 	}
 
 	return(0);
@@ -7324,10 +7324,10 @@
 		transaction */
 		if (trx->active_trans == 0) {
 
-			innobase_register_trx_and_stmt(engine, session);
+			innobase_register_trx_and_stmt(innodb_engine_ptr, session);
 			trx->active_trans = 1;
 		} else if (trx->n_mysql_tables_in_use == 0) {
-			innobase_register_stmt(engine, session);
+			innobase_register_stmt(innodb_engine_ptr, session);
 		}
 
 		if (trx->isolation_level == TRX_ISO_SERIALIZABLE
@@ -7389,7 +7389,7 @@
 
 		if (!session_test_options(session, OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)) {
 			if (trx->active_trans != 0) {
-				engine->commit(session, TRUE);
+				getTransactionalEngine()->commit(session, TRUE);
 			}
 		} else {
 			if (trx->isolation_level <= TRX_ISO_READ_COMMITTED
@@ -8249,7 +8249,7 @@
 This function is used to prepare an X/Open XA distributed transaction.
 @return	0 or error number */
 int
-InnobaseEngine::prepare(
+InnobaseEngine::doPrepare(
 /*================*/
 	Session*	session,/*!< in: handle to the MySQL thread of
 				the user whose XA transaction should
@@ -8348,10 +8348,10 @@
 This function is used to recover X/Open XA distributed transactions.
 @return	number of prepared transactions stored in xid_list */
 int
-InnobaseEngine::recover(
+InnobaseEngine::doRecover(
 /*================*/
 	::drizzled::XID*	xid_list,/*!< in/out: prepared transactions */
-	uint			len)	/*!< in: number of slots in xid_list */
+	size_t len)	/*!< in: number of slots in xid_list */
 {
 	assert(this == innodb_engine_ptr);
 
@@ -8368,7 +8368,7 @@
 which is in the prepared state
 @return	0 or error number */
 int
-InnobaseEngine::commit_by_xid(
+InnobaseEngine::doCommitXid(
 /*===================*/
 	::drizzled::XID*	xid)	/*!< in: X/Open XA transaction identification */
 {
@@ -8392,7 +8392,7 @@
 which is in the prepared state
 @return	0 or error number */
 int
-InnobaseEngine::rollback_by_xid(
+InnobaseEngine::doRollbackXid(
 /*=====================*/
 	::drizzled::XID*		xid)	/*!< in: X/Open XA transaction
 				identification */

=== modified file 'plugin/innobase/handler/ha_innodb.h'
--- plugin/innobase/handler/ha_innodb.h	2010-02-04 08:14:46 +0000
+++ plugin/innobase/handler/ha_innodb.h	2010-02-14 19:27:57 +0000
@@ -28,6 +28,7 @@
 
 #include <drizzled/cursor.h>
 #include <drizzled/thr_lock.h>
+#include <drizzled/plugin/transactional_storage_engine.h>
 
 using namespace drizzled;
 /** InnoDB table share */
@@ -102,6 +103,19 @@
 	UNIV_INTERN ha_innobase(plugin::StorageEngine &engine,
                                 TableShare &table_arg);
 	UNIV_INTERN ~ha_innobase();
+  /**
+   * Returns the plugin::TransactionStorageEngine pointer
+   * of the cursor's underlying engine.
+   *
+   * @todo
+   *
+   * Have a TransactionalCursor subclass...
+   */
+  UNIV_INTERN plugin::TransactionalStorageEngine *getTransactionalEngine()
+  {
+    return static_cast<plugin::TransactionalStorageEngine *>(engine);
+  }
+
 	/*
 	  Get the row type from the storage engine.  If this method returns
 	  ROW_TYPE_NOT_USED, the information in HA_CREATE_INFO should be used.

